from builtins import object

from copy import deepcopy


class DeprecationError(Exception):
    pass


class NetCDF(object):
    '''Mixin class for storing simple netCDF elements.

.. versionadded:: 1.7.0

    '''
    def _initialise_netcdf(self, source=None):
        '''Call this from inside the __init__ method of a class that inherits
    from this mixin class.
    
    :Parameters:
    
        source: optional
            Initialise the netCDF components from those of *source*.
    
    :Returns:
    
        `None`
    
    **Examples:**
    
    >>> f._initialise_netcdf(source)

        '''
        if source is None:
             netcdf = {}
        else:        
            try:
                netcdf = source._get_component('netcdf', {})
            except AttributeError:
                netcdf = {}
            else:
                if netcdf:
                    netcdf = deepcopy(netcdf)
                else:
                    netcdf = {}
        #--- End: if
        
        self._set_component('netcdf', netcdf, copy=False)


#--- End: class


class NetCDFDimension(NetCDF):
    '''Mixin class for accessing the netCDF dimension name.

.. versionadded:: 1.7.0

    '''
    def nc_del_dimension(self, default=ValueError()):
        '''Remove the netCDF dimension name.

    .. versionadded:: 1.7.0
    
    .. seealso:: `nc_get_dimension`, `nc_has_dimension`,
                 `nc_set_dimension`
    
    :Parameters:
    
        default: optional
            Return the value of the *default* parameter if the netCDF
            dimension name has not been set. If set to an `Exception`
            instance then it will be raised instead.
    
    :Returns:
    
        `str`
            The removed netCDF dimension name.
    
    **Examples:**
    
    >>> f.nc_set_dimension('time')
    >>> f.nc_has_dimension()
    True
    >>> f.nc_get_dimension()
    'time'
    >>> f.nc_del_dimension()
    'time'
    >>> f.nc_has_dimension()
    False
    >>> print(f.nc_get_dimension(None))
    None
    >>> print(f.nc_del_dimension(None))
    None

        '''
        try:
            return self._get_component('netcdf').pop('dimension')
        except KeyError:
            return self._default(default,
                   "{!r} has no netCDF dimension name".format(
                       self.__class__.__name__))


    def nc_get_dimension(self, default=ValueError()):
        '''Return the netCDF dimension name.

    .. versionadded:: 1.7.0
    
    .. seealso:: `nc_del_dimension`, `nc_has_dimension`,
                 `nc_set_dimension`
    
    :Parameters:
    
        default: optional
            Return the value of the *default* parameter if the netCDF
            dimension name has not been set. If set to an `Exception`
            instance then it will be raised instead.
    
    :Returns:
    
        `str`
            The netCDF dimension name.
    
    **Examples:**
    
    >>> f.nc_set_dimension('time')
    >>> f.nc_has_dimension()
    True
    >>> f.nc_get_dimension()
    'time'
    >>> f.nc_del_dimension()
    'time'
    >>> f.nc_has_dimension()
    False
    >>> print(f.nc_get_dimension(None))
    None
    >>> print(f.nc_del_dimension(None))
    None

        '''   
        try:
            return self._get_component('netcdf')['dimension']
        except KeyError:
            return self._default(default,
                   "{!r} has no netCDF dimension name".format(
                       self.__class__.__name__))


    def nc_has_dimension(self):
        '''Whether the netCDF dimension name has been set.

    .. versionadded:: 1.7.0
    
    .. seealso:: `nc_del_dimension`, `nc_get_dimension`,
                 `nc_set_dimension`
    
    :Returns:
    
        `bool`
            `True` if the netCDF dimension name has been set,
            otherwise `False`.
    
    **Examples:**
    
    >>> f.nc_set_dimension('time')
    >>> f.nc_has_dimension()
    True
    >>> f.nc_get_dimension()
    'time'
    >>> f.nc_del_dimension()
    'time'
    >>> f.nc_has_dimension()
    False
    >>> print(f.nc_get_dimension(None))
    None
    >>> print(f.nc_del_dimension(None))
    None

        '''
        return 'dimension' in self._get_component('netcdf')


    def nc_set_dimension(self, value):
        '''Set the netCDF dimension name.
        
    .. versionadded:: 1.7.0
    
    .. seealso:: `nc_del_dimension`, `nc_get_dimension`,
                 `nc_has_dimension`
    
    :Parameters:
    
        value: `str`
            The value for the netCDF dimension name.
    
    :Returns:
    
        `None`
    
    **Examples:**
    
    >>> f.nc_set_dimension('time')
    >>> f.nc_has_dimension()
    True
    >>> f.nc_get_dimension()
    'time'
    >>> f.nc_del_dimension()
    'time'
    >>> f.nc_has_dimension()
    False
    >>> print(f.nc_get_dimension(None))
    None
    >>> print(f.nc_del_dimension(None))
    None

        '''
        self._get_component('netcdf')['dimension'] = value


#--- End: class


class NetCDFVariable(NetCDF):
    '''Mixin class for accessing the netCDF variable name.

.. versionadded:: 1.7.0

    '''
    def nc_del_variable(self, default=ValueError()):
        '''Remove the netCDF variable name.

    .. versionadded:: 1.7.0
    
    .. seealso:: `nc_get_variable`, `nc_has_variable`,
                 `nc_set_variable`
    
    :Parameters:
    
        default: optional
            Return the value of the *default* parameter if the netCDF
            variable name has not been set. If set to an `Exception`
            instance then it will be raised instead.
    
    :Returns:
    
        `str`
            The removed netCDF variable name.
    
    **Examples:**
    
    >>> f.nc_set_variable('tas')
    >>> f.nc_has_variable()
    True
    >>> f.nc_get_variable()
    'tas'
    >>> f.nc_del_variable()
    'tas'
    >>> f.nc_has_variable()
    False
    >>> print(f.nc_get_variable(None))
    None
    >>> print(f.nc_del_variable(None))
    None

        '''        
        try:
            return self._get_component('netcdf').pop('variable')
        except KeyError:
            return self._default(default,
                   "{!r} has no netCDF variable name".format(
                       self.__class__.__name__))

        
    def nc_get_variable(self, default=ValueError()):
        '''Return the netCDF variable name.

    .. versionadded:: 1.7.0
    
    .. seealso:: `nc_del_variable`, `nc_has_variable`,
                 `nc_set_variable`
    
    :Parameters:
    
        default: optional
            Return the value of the *default* parameter if the netCDF
            variable name has not been set. If set to an `Exception`
            instance then it will be raised instead.
    
    :Returns:
    
        `str`
            The netCDF variable name. If unset then *default* is
            returned, if provided.
    
    **Examples:**
    
    >>> f.nc_set_variable('tas')
    >>> f.nc_has_variable()
    True
    >>> f.nc_get_variable()
    'tas'
    >>> f.nc_del_variable()
    'tas'
    >>> f.nc_has_variable()
    False
    >>> print(f.nc_get_variable(None))
    None
    >>> print(f.nc_del_variable(None))
    None

        '''
        try:
            return self._get_component('netcdf')['variable']
        except KeyError:
            return self._default(default,
                   "{!r} has no netCDF variable name".format(
                       self.__class__.__name__))


    def nc_has_variable(self):
        '''Whether the netCDF variable name has been set.

    .. versionadded:: 1.7.0
    
    .. seealso:: `nc_del_variable`, `nc_get_variable`,
                 `nc_set_variable`
    
    :Returns:
    
        `bool`
            `True` if the netCDF variable name has been set, otherwise
            `False`.
    
    **Examples:**
    
    >>> f.nc_set_variable('tas')
    >>> f.nc_has_variable()
    True
    >>> f.nc_get_variable()
    'tas'
    >>> f.nc_del_variable()
    'tas'
    >>> f.nc_has_variable()
    False
    >>> print(f.nc_get_variable(None))
    None
    >>> print(f.nc_del_variable(None))
    None

        '''
        return 'variable' in self._get_component('netcdf')


    def nc_set_variable(self, value):
        '''Set the netCDF variable name.

    .. versionadded:: 1.7.0
    
    .. seealso:: `nc_del_variable`, `nc_get_variable`,
                 `nc_has_variable`
    
    :Parameters:
    
        value: `str`
            The value for the netCDF variable name.
    
    :Returns:
    
        `None`
    
    **Examples:**
    
    >>> f.nc_set_variable('tas')
    >>> f.nc_has_variable()
    True
    >>> f.nc_get_variable()
    'tas'
    >>> f.nc_del_variable()
    'tas'
    >>> f.nc_has_variable()
    False
    >>> print(f.nc_get_variable(None))
    None
    >>> print(f.nc_del_variable(None))
    None

        '''
        self._get_component('netcdf')['variable'] = value


#--- End: class


class NetCDFSampleDimension(NetCDF):
    '''Mixin class for accessing the netCDF sample dimension name.

.. versionadded:: 1.7.0

    '''
    def nc_del_sample_dimension(self, default=ValueError()):
        '''Remove the netCDF sample dimension name.

    .. versionadded:: 1.7.0
    
    .. seealso:: `nc_get_sample_dimension`, `nc_has_sample_dimension`,
                 `nc_set_sample_dimension`
    
    :Parameters:
    
        default: optional
            Return the value of the *default* parameter if the netCDF
            sample dimension name has not been set. If set to an
            `Exception` instance then it will be raised instead.
    
    :Returns:
    
        `str`
            The removed netCDF sample dimension name.
    
    **Examples:**
    
    >>> f.nc_set_sample_dimension('time')
    >>> f.nc_has_sample_dimension()
    True
    >>> f.nc_get_sample_dimension()
    'time'
    >>> f.nc_del_sample_dimension()
    'time'
    >>> f.nc_has_sample_dimension()
    False
    >>> print(f.nc_get_sample_dimension(None))
    None
    >>> print(f.nc_del_sample_dimension(None))
    None

        '''
        try:
            return self._get_component('netcdf').pop('sample_dimension')
        except KeyError:
            return self._default(default,
                   "{!r} has no netCDF sample dimension name".format(
                       self.__class__.__name__))


    def nc_get_sample_dimension(self, default=ValueError()):
        '''Return the netCDF sample dimension name.

    .. versionadded:: 1.7.0
    
    .. seealso:: `nc_del_sample_dimension`, `nc_has_sample_dimension`,
                 `nc_set_sample_dimension`
    
    :Parameters:
    
        default: optional
            Return the value of the *default* parameter if the netCDF
            sample dimension name has not been set. If set to an
            `Exception` instance then it will be raised instead.
    
    :Returns:
    
        `str`
            The netCDF sample dimension name.
    
    **Examples:**
    
    >>> f.nc_set_sample_dimension('time')
    >>> f.nc_has_sample_dimension()
    True
    >>> f.nc_get_sample_dimension()
    'time'
    >>> f.nc_del_sample_dimension()
    'time'
    >>> f.nc_has_sample_dimension()
    False
    >>> print(f.nc_get_sample_dimension(None))
    None
    >>> print(f.nc_del_sample_dimension(None))
    None

        '''
        try:
            return self._get_component('netcdf')['sample_dimension']
        except KeyError:
            return self._default(default,
                   "{!r} has no netCDF sample dimension name".format(
                       self.__class__.__name__))


    def nc_has_sample_dimension(self):
        '''Whether the netCDF sample dimension name has been set.

    .. versionadded:: 1.7.0
    
    .. seealso:: `nc_del_sample_dimension`, `nc_get_sample_dimension`,
                 `nc_set_sample_dimension`
    
    :Returns:
         
        `bool`
            `True` if the netCDF sample dimension name has been set,
            otherwise `False`.
    
    **Examples:**
    
    >>> f.nc_set_sample_dimension('time')
    >>> f.nc_has_sample_dimension()
    True
    >>> f.nc_get_sample_dimension()
    'time'
    >>> f.nc_del_sample_dimension()
    'time'
    >>> f.nc_has_sample_dimension()
    False
    >>> print(f.nc_get_sample_dimension(None))
    None
    >>> print(f.nc_del_sample_dimension(None))
    None

        '''
        return 'sample_dimension' in self._get_component('netcdf')


    def nc_set_sample_dimension(self, value):
        '''Set the netCDF sample dimension name.

    .. versionadded:: 1.7.0
    
    .. seealso:: `nc_del_sample_dimension`, `nc_get_sample_dimension`,
                 `nc_has_sample_dimension`
                 
    :Parameters:
    
        value: `str`
            The value for the netCDF sample dimension name.
    
    :Returns:
    
        `None`
    
    **Examples:**
    
    >>> f.nc_set_sample_dimension('time')
    >>> f.nc_has_sample_dimension()
    True
    >>> f.nc_get_sample_dimension()
    'time'
    >>> f.nc_del_sample_dimension()
    'time'
    >>> f.nc_has_sample_dimension()
    False
    >>> print(f.nc_get_sample_dimension(None))
    None
    >>> print(f.nc_del_sample_dimension(None))
    None

        '''
        self._get_component('netcdf')['sample_dimension'] = value


#--- End: class


class NetCDFGlobalAttributes(NetCDF):
    '''Mixin class for accessing netCDF global attributes.

.. versionadded:: 1.7.0

    '''
    def nc_global_attributes(self):
        '''Return the selection of properties to be written as netCDF global
    attributes.
    
    When multiple field constructs are being written to the same file,
    it is only possible to create a netCDF global attribute from a
    property that has identical values for each field construct. If
    any field construct's property has a different value then the
    property will not be written as a netCDF global attribute, even if
    it has been selected as such, but will appear instead as
    attributes on the netCDF data variables corresponding to each
    field construct.
    
    The standard description-of-file-contents properties are always
    written as netCDF global attributes, if possible, so selecting
    them is optional.
    
    .. versionadded:: 1.7.0
    
    .. seealso:: `cfdm.write`, `nc_clear_global_attributes`,
                 `nc_set_global_attribute`
    
    :Returns:
    
        `dict`
            The selection of properties requested for writting to
            netCDF global attributes.
    
    **Examples:**
    
    >>> f.nc_global_attributes()
    {'Conventions': None, 'comment': None}
    >>> f.nc_set_global_attribute('foo')
    >>> f.nc_global_attributes()
    {'Conventions': None, 'comment': None, 'foo'}
    >>> f.nc_set_global_attribute('comment', 'global comment')
    >>> f.nc_global_attributes()
    {'Conventions': None, 'comment': 'global_comment', 'foo'}
    >>> f.nc_clear_global_attributes()
    {'Conventions': None, 'comment': 'global_comment', 'foo'}
    >>> f.nc_global_attributes()
    {}

        '''
        out = self._get_component('netcdf').get('global_attributes')
        
        if out is None:
            return {}
        
        return out.copy()

    
    def nc_clear_global_attributes(self):
        '''Remove the selection of properties to be written as netCDF global
    attributes.
    
    When multiple field constructs are being written to the same file,
    it is only possible to create a netCDF global attribute from a
    property that has identical values for each field construct. If
    any field construct's property has a different value then the
    property will not be written as a netCDF global attribute, even if
    it has been selected as such, but will appear instead as
    attributes on the netCDF data variables corresponding to each
    field construct.
    
    The standard description-of-file-contents properties are always
    written as netCDF global attributes, if possible, so selecting
    them is optional.
    
    .. versionadded:: 1.7.0
    
    .. seealso:: `cfdm.write`, `nc_global_attributes`,
                 `nc_set_global_attribute`
    
    :Returns:
    
        `dict`
            The removed selection of properties requested for writting
            to netCDF global attributes.
    
    **Examples:**
    
    >>> f.nc_global_attributes()
    {'Conventions': None, 'comment': None}
    >>> f.nc_set_global_attribute('foo')
    >>> f.nc_global_attributes()
    {'Conventions': None, 'comment': None, 'foo'}
    >>> f.nc_set_global_attribute('comment', 'global comment')
    >>> f.nc_global_attributes()
    {'Conventions': None, 'comment': 'global_comment', 'foo'}
    >>> f.nc_clear_global_attributes()
    {'Conventions': None, 'comment': 'global_comment', 'foo'}
    >>> f.nc_global_attributes()
    {}

        '''
        out = self._get_component('netcdf').get('global_attributes')
        
        if out is None:
            out = {}

        self._get_component('netcdf')['global_attributes'] = {}

        return out

    
    def nc_set_global_attribute(self, prop, value=None):
        '''Select a property to be written as a netCDF global attribute.

    When multiple field constructs are being written to the same file,
    it is only possible to create a netCDF global attribute from a
    property that has identical values for each field construct. If
    any field construct's property has a different value then the
    property will not be written as a netCDF global attribute, even if
    it has been selected as such, but will appear instead as
    attributes on the netCDF data variables corresponding to each
    field construct.
    
    The standard description-of-file-contents properties are always
    written as netCDF global attributes, if possible, so selecting
    them is optional.
    
    .. versionadded:: 1.7.0
    
    .. seealso:: `cfdm.write`, `nc_global_attributes`,
                 `nc_clear_global_attributes`
    
    :Parameters:
    
        prop: `str`
            Select the property to be written (if possible) as a
            netCDF global attribute.
    
        value: optional
            The value of the netCDF global attribute, which will be
            created (if possible) in addition to the property as
            written to a netCDF data variable. If unset (or `None`)
            then this acts as an instruction to write the property (if
            possible) to a netCDF global attribute instead of to a
            netCDF data variable.
            
    :Returns:
    
        `None`
    
    **Examples:**
    
    >>> f.nc_global_attributes()
    {'Conventions': None, 'comment': None}
    >>> f.nc_set_global_attribute('foo')
    >>> f.nc_global_attributes()
    {'Conventions': None, 'comment': None, 'foo'}
    >>> f.nc_set_global_attribute('comment', 'global comment')
    >>> f.nc_global_attributes()
    {'Conventions': None, 'comment': 'global_comment', 'foo'}
    >>> f.nc_clear_global_attributes()
    {'Conventions': None, 'comment': 'global_comment', 'foo'}
    >>> f.nc_global_attributes()
    {}

        '''
        out = self._get_component('netcdf').get('global_attributes')
        
        if out is None:
            out = {}        

        out[prop] = value
        
        self._get_component('netcdf')['global_attributes'] = out

    
#--- End: class


class NetCDFUnlimitedDimensions(NetCDF):
    '''Mixin class for accessing netCDF unlimited dimensions.

.. versionadded:: 1.7.0

Deprecated at version 1.7.4

    '''
    def nc_unlimited_dimensions(self):
        '''Return the selection of domain axis constructs to be written as
    netCDF unlimited dimensions.
    
    By default output netCDF dimensions are not unlimited.
    
    .. versionadded:: 1.7.0
    
    Deprecated at version 1.7.4
    
    .. seealso:: `cfdm.write`, `nc_clear_unlimited_dimensions`,
                 `nc_set_unlimited_dimensions`
    
    :Returns:
    
        `set`
            The selection of domain axis constructs to be written as
            netCDF unlimited dimensions.
    
    **Examples:**
    
    >>> f.nc_set_unlimited_dimensions(['domainaxis0'])
    >>> f.nc_unlimited_dimensions()
    {'domainaxis0'}
    >>> f.nc_set_unlimited_dimensions(['domainaxis1'])
    >>> f.nc_unlimited_dimensions()
    {'domainaxis0', 'domainaxis1'}
    >>> f.nc_clear_unlimited_dimensions()
    {'domainaxis0', 'domainaxis1'}
    >>> f.nc_unlimited_dimensions()
    set()

        '''
        raise DeprecationError("Field.nc_unlimited_dimensions was deprecated at v1.7.4 and is no longer available. Use DomainAxis.nc_is_unlimited instead")
    
        out = self._get_component('netcdf').get('unlimited_dimensions')
        
        if out is None:
            return set()

        return set(out)


    def nc_set_unlimited_dimensions(self, axes):
        '''Select domain axis constructs to be written as netCDF unlimited
    dimensions.
    
    By default output netCDF dimensions are not unlimited.
    
    .. versionadded:: 1.7.0
    
    Deprecated at version 1.7.4
    
    .. seealso:: `cfdm.write`, `nc_unlimited_dimensions`,
                 `nc_clear_unlimited_dimensions`
    
    :Parameters:
    
        axes: sequence of `str`, optional   
            Select the domain axis constructs from the sequence
            provided. Domain axis constructs are identified by their
            construct identifiers.
    
            *Parameter example:*
              ``axes=['domainaxis0', 'domainaxis1']``
            
            *Parameter example:*
              ``axes=()``        
    
    :Returns:
    
        `None`
    
    **Examples:**
    
    >>> f.nc_set_unlimited_dimensions(['domainaxis0'])
    >>> f.nc_unlimited_dimensions()
    {'domainaxis0'}
    >>> f.nc_set_unlimited_dimensions(['domainaxis1'])
    >>> f.nc_unlimited_dimensions()
    {'domainaxis0', 'domainaxis1'}
    >>> f.nc_clear_unlimited_dimensions()
    {'domainaxis0', 'domainaxis1'}
    >>> f.nc_unlimited_dimensions()
    set()

        '''
        raise DeprecationError("Field.nc_set_unlimited_dimensions was deprecated at v1.7.4  and is no longer available. Use DomainAxis.nc_set_unlimited instead")
                
        out = self._get_component('netcdf').get('unlimited_dimensions')
        
        if out is None:
            out = set()
        else:
            out = set(out)

        out.update(axes)
            
        self._get_component('netcdf')['unlimited_dimensions'] = tuple(out)


    def nc_clear_unlimited_dimensions(self):
        '''Remove the selection of domain axis constructs to be written as
    netCDF unlimited dimensions.
    
    By default output netCDF dimensions are not unlimited.
    
    .. versionadded:: 1.7.0
    
    Deprecated at version 1.7.4
    
    .. seealso:: `cfdm.write`, `nc_unlimited_dimensions`,
                 `nc_set_unlimited_dimensions`
    
    :Returns:
    
        `set`
            The selection of domain axis constructs that has been removed.
    
    **Examples:**
    
    >>> f.nc_set_unlimited_dimensions(['domainaxis0'])
    >>> f.nc_unlimited_dimensions()
    {'domainaxis0'}
    >>> f.nc_set_unlimited_dimensions(['domainaxis1'])
    >>> f.nc_unlimited_dimensions()
    {'domainaxis0', 'domainaxis1'}
    >>> f.nc_clear_unlimited_dimensions()
    {'domainaxis0', 'domainaxis1'}
    >>> f.nc_unlimited_dimensions()
    set()

        '''
        raise DeprecationError("Field.nc_clear_unlimited_dimensions was deprecated at v1.7.4 and is no longer available. Use DomainAxis.nc_set_unlimited instead")                

        out = self._get_component('netcdf').get('unlimited_dimensions')
        
        if out is None:
            out = set()
        else:
            out = set(out)

        self._get_component('netcdf')['unlimited_dimensions'] = ()

        return out

    
#--- End: class


class NetCDFExternal(NetCDF):
    '''Mixin class for accessing the netCDF external variable status.

.. versionadded:: 1.7.0

    '''
    def nc_get_external(self):
        '''Whether the construct corresponds to an external netCDF variable.

    .. versionadded:: 1.7.0
    
    .. seealso:: `nc_set_external`
    
    :Returns:
    
        `bool`
            The external status.
    
    **Examples:**
    
    >>> c.nc_get_external()
    False
    >>> c.nc_set_external(True)
    >>> c.nc_get_external()
    True

        '''
        return self._get_component('netcdf').get('external', False)


    def nc_set_external(self, external):
        '''Set external status of a netCDF variable.

    .. versionadded:: 1.7.0
    
    .. seealso:: `nc_get_external`
    
    :Parameters:
    
        external: `bool`, optional
            Set the external status.
    
            *Parameter example:*
              ``external=True``
    
    :Returns:
    
        `None`
    
    **Examples:**
    
    >>> c.nc_get_external()
    False
    >>> c.nc_set_external(True)
    >>> c.nc_get_external()
    True

        '''
        self._get_component('netcdf')['external'] = bool(external)


#--- End: class

class NetCDFGeometry(NetCDF):
    '''Mixin class for accessing the netCDF geometry container variable
name.

.. versionadded:: 1.8.0

    '''   
    def nc_del_geometry_variable(self, default=ValueError()):
        '''Remove the netCDF geometry container variable name.

    .. versionadded:: 1.8.0

    .. seealso:: `nc_get_geometry_variable`,
                 `nc_has_geometry_variable`,
                 `nc_set_geometry_variable`

    :Parameters:

        default: optional
            Return the value of the *default* parameter if the netCDF
            dimension name has not been set. If set to an `Exception`
            instance then it will be raised instead.

    :Returns:

        `str`
            The removed netCDF geometry container variable name.

    **Examples:**

    >>> f.nc_set_geometry_variable('geometry')
    >>> f.nc_has_geometry_variable()
    True
    >>> f.nc_get_geometry_variable()
    'geometry'
    >>> f.nc_del_geometry_variable()
    'geometry'
    >>> f.nc_has_geometry_variable()
    False
    >>> print(f.nc_get_geometry_variable(None))
    None
    >>> print(f.nc_del_geometry_variable(None))
    None

        '''
        try:
            return self._get_component('netcdf').pop('geometry_variable')
        except KeyError:
            return self._default(default,
                   "{!r} has no netCDF geometry variable name".format(
                       self.__class__.__name__))


    def nc_get_geometry_variable(self, default=ValueError()):
        '''Return the netCDF geometry container variable name.

    .. versionadded:: 1.8.0

    .. seealso:: `nc_del_geometry_variable`,
                 `nc_has_geometry_variable`,
                 `nc_set_geometry_variable`

    :Parameters:

        default: optional
            Return the value of the *default* parameter if the netCDF
            dimension name has not been set. If set to an `Exception`
            instance then it will be raised instead.

    :Returns:

        `str`
            The netCDF geometry container variable name.

    **Examples:**

    >>> f.nc_set_geometry_variable('geometry')
    >>> f.nc_has_geometry_variable()
    True
    >>> f.nc_get_geometry_variable()
    'geometry'
    >>> f.nc_del_geometry_variable()
    'geometry'
    >>> f.nc_has_geometry_variable()
    False
    >>> print(f.nc_get_geometry_variable(None))
    None
    >>> print(f.nc_del_geometry_variable(None))
    None

        '''   
        try:
            return self._get_component('netcdf')['geometry_variable']
        except KeyError:
            return self._default(default,
                   "{!r} has no netCDF geometry variable name".format(
                       self.__class__.__name__))


    def nc_has_geometry_variable(self):
        '''Whether the netCDF geometry container variable name has been set.

    .. versionadded:: 1.8.0

    .. seealso:: `nc_del_geometry_variable`,
                 `nc_get_geometry_variable`,
                 `nc_set_geometry_variable`

    :Returns:
        `bool`
            `True` if the netCDF geometry container variable name has
            been set, otherwise `False`.

    **Examples:**

    >>> f.nc_set_geometry_variable('geometry')
    >>> f.nc_has_geometry_variable()
    True
    >>> f.nc_get_geometry_variable()
    'geometry'
    >>> f.nc_del_geometry_variable()
    'geometry'
    >>> f.nc_has_geometry_variable()
    False
    >>> print(f.nc_get_geometry_variable(None))
    None
    >>> print(f.nc_del_geometry_variable(None))
    None

        '''
        return 'geometry_variable' in self._get_component('netcdf')

    
    def nc_set_geometry_variable(self, value):
        '''Set the netCDF geometry container variable name.
        
    .. versionadded:: 1.8.0
    
    .. seealso:: `nc_del_geometry_variable`,
                 `nc_get_geometry_variable`,
                 `nc_has_geometry_variable`

    :Parameters:

        value: `str`
            The value for the netCDF geometry container variable name.

    :Returns:
        `None`

    **Examples:**

    >>> f.nc_set_geometry_variable('geometry')
    >>> f.nc_has_geometry_variable()
    True
    >>> f.nc_get_geometry_variable()
    'geometry'
    >>> f.nc_del_geometry_variable()
    'geometry'
    >>> f.nc_has_geometry_variable()
    False
    >>> print(f.nc_get_geometry_variable(None))
    None
    >>> print(f.nc_del_geometry_variable(None))
    None

        '''
        self._get_component('netcdf')['geometry_variable'] = value


#--- End: class

class NetCDFHDF5(NetCDF):
    '''Mixin class for TODO

.. versionadded:: 1.7.2

    '''
    def nc_hdf5_chunksizes(self):
        '''TODO

    .. note:: Chunksizes are cleared from the output of methods that
              change the data shape.
    
    .. note:: Chunksizes are ignored for netCDF3 files that do not use
              HDF5.

.. versionadded:: 1.7.2
    
    .. seealso:: `nc_clear_hdf5_chunksizes`, `nc_set_hdf5_chunksizes`
    
    :Returns:
    
        `tuple`
            TODO The chunk sizes prior to the new setting, or the current
            current sizes if no new values are specified.
    
    **Examples:**
    
    >>> d.shape
    (1, 96, 73)
    >>> d.nc_set_hdf5_chunksizes([1, 48, 73])
    >>> d.nc_hdf5_chunksizes()
    (1, 48, 73)
    >>> d.nc_clear_hdf5_chunksizes()
    (1, 48, 73)
    >>> d.nc_hdf5_chunksizes()
    ()

        '''
        return self._get_component('netcdf').get('hdf5_chunksizes', ())


    def nc_clear_hdf5_chunksizes(self):
        '''TODO

    .. note:: Chunksizes are cleared from the output of methods that
              change the data shape.
    
    .. note:: Chunksizes are ignored for netCDF3 files that do not use
              HDF5.

    .. versionadded:: 1.7.2
    
    .. seealso:: `nc_hdf5_chunksizes`, `nc_set_hdf5_chunksizes`
    
    :Returns:
    
        `tuple`
            TODO
    
    **Examples:**
    
    >>> d.shape
    (1, 96, 73)
    >>> d.nc_set_hdf5_chunksizes([1, 48, 73])
    >>> d.nc_hdf5_chunksizes()
    (1, 48, 73)
    >>> d.nc_clear_hdf5_chunksizes()
    (1, 48, 73)
    >>> d.nc_hdf5_chunksizes()
    ()

        '''
        return self._get_component('netcdf').pop('hdf5_chunksizes', ())


    def nc_set_hdf5_chunksizes(self, chunksizes):
        '''TODO

    .. note:: Chunksizes are cleared from the output of methods that
              change the data shape.
   
    .. note:: Chunksizes are ignored for netCDF3 files that do not use
              HDF5.

    .. versionadded:: 1.7.2
    
    .. seealso:: `nc_hdf5_chunksizes`, `nc_clear_hdf5_chunksizes`
    
    :Parameters:
    
        chunksizes: sequence of `int`
            The chunksizes for each dimension. Can be integers from 0
            to the dimension size.
    
    :Returns:
    
        `None`
    
    **Examples:**
    
    >>> d.shape
    (1, 96, 73)
    >>> d.nc_set_hdf5_chunksizes([1, 48, 73])
    >>> d.nc_hdf5_chunksizes()
    (1, 48, 73)
    >>> d.nc_clear_hdf5_chunksizes()
    (1, 48, 73)
    >>> d.nc_hdf5_chunksizes()
    ()

        '''
        try:
            shape = self.shape
        except AttributeError:
            pass
        else:
            if len(chunksizes) != len(shape):
                raise ValueError(
"chunksizes must be a sequence with the same length as dimensions")

            for i, j in zip(chunksizes, shape):
                if i < 0:
                    raise ValueError("chunksize cannot be negative")
                if i > j:
                    raise ValueError("chunksize cannot exceed dimension size")
        #--- End: try
        
        self._get_component('netcdf')['hdf5_chunksizes'] = tuple(chunksizes)


#--- End: class

class NetCDFHDF5_exp(NetCDF):
    '''Mixin class for TODO

.. versionadded:: 1.7.2

    '''
    def nc_del_hdf5_chunksize(self, default=ValueError()):
        '''TODO

    .. note:: Chunksizes are cleared from the output of methods that
              change the data shape, and chunksizes are ignored for
              netCDF3 files that do not use HDF5.

    .. versionadded:: 1.7.4
    
    .. seealso:: `nc_get_hdf5_chunksize`, `nc_has_hdf5_chunksize`,
                 `nc_set_hdf5_chunksize`
    
    :Returns:
    
            TODO
    
    **Examples:**
    
    TODO
    
    >>> d.shape
    (1, 96, 73)
    >>> d.nc_set_hdf5_chunksizes([1, 48, 73])
    >>> d.nc_hdf5_chunksizes()
    (1, 48, 73)
    >>> d.nc_clear_hdf5_chunksizes()
    (1, 48, 73)
    >>> d.nc_hdf5_chunksizes()
    ()

        '''
        return 'hdf5_chunksize' in self._get_component('netcdf')


    def nc_get_hdf5_chunksize(self, default=ValueError()):
        '''TODO

    .. note:: Chunksizes are cleared from the output of methods that
              change the data shape, and chunksizes are ignored for
              netCDF3 files that do not use HDF5.

    .. versionadded:: 1.7.4
    
    .. seealso:: `nc_del_hdf5_chunksize`, `nc_has_hdf5_chunksize`,
                 `nc_set_hdf5_chunksize`
    
    :Returns:
    
            TODO The chunk sizes prior to the new setting, or the current
            current sizes if no new values are specified.
    
    **Examples:**
    
    TODO
    
    >>> d.shape
    (1, 96, 73)
    >>> d.nc_set_hdf5_chunksizes([1, 48, 73])
    >>> d.nc_hdf5_chunksizes()
    (1, 48, 73)
    >>> d.nc_clear_hdf5_chunksizes()
    (1, 48, 73)
    >>> d.nc_hdf5_chunksizes()
    ()

        '''
        try:
            return self._get_component('netcdf')['hdf5_chunksize']
        except KeyError:
            return self._default(default,
                   "{!r} has no HDF5 chunksize".format(self.__class__.__name__))


    def nc_has_hdf5_chunksize(self):
        '''TODO

    .. note:: Chunksizes are cleared from the output of methods that
              change the data shape, and chunksizes are ignored for
              netCDF3 files that do not use HDF5.

    .. versionadded:: 1.7.4
    
    .. seealso:: `nc_del_hdf5_chunksize`, `nc_get_hdf5_chunksize`,
                 `nc_set_hdf5_chunksize`
    
    :Returns:
    
        `bool`
            TODO The chunk sizes prior to the new setting, or the current
            current sizes if no new values are specified.
    
    **Examples:**
    
    TODO
    
    >>> d.shape
    (1, 96, 73)
    >>> d.nc_set_hdf5_chunksizes([1, 48, 73])
    >>> d.nc_hdf5_chunksizes()
    (1, 48, 73)
    >>> d.nc_clear_hdf5_chunksizes()
    (1, 48, 73)
    >>> d.nc_hdf5_chunksizes()
    ()

        '''
        return 'hdf5_chunksize' in self._get_component('netcdf')


    def nc_set_hdf5_chunksize(self, value):
        '''TODO

    .. note:: Chunksizes are cleared from the output of methods that
              change the data shape, and chunksizes are ignored for
              netCDF3 files that do not use HDF5.

    .. versionadded:: 1.7.4
    
    .. seealso:: `nc_del_hdf5_chunksize`, `nc_get_hdf5_chunksize`,
                 `nc_has_hdf5_chunksize`
    
    :Parameters:
    
        value: `int`
            TODO
    
    :Returns:
    
        `None`
    
    **Examples:**
    
    TODO
    
    >>> d.shape
    (1, 96, 73)
    >>> d.nc_set_hdf5_chunksizes([1, 48, 73])
    >>> d.nc_hdf5_chunksizes()
    (1, 48, 73)
    >>> d.nc_clear_hdf5_chunksizes()
    (1, 48, 73)
    >>> d.nc_hdf5_chunksizes()
    ()

        '''
        self._get_component('netcdf')['hdf5_chunksize'] = int(value)


#--- End: class

class NetCDFUnlimitedDimension(NetCDF):
    '''Mixin class for accessing a netCDF unlimited dimension.

.. versionadded:: 1.7.4

    '''
    def nc_is_unlimited(self):
        '''Inspect the unlimited status of the a netCDF dimension.

    By default output netCDF dimensions are not unlimited. The status
    is used by the `write` function.
    
    .. versionadded:: 1.7.4
    
    .. seealso:: `nc_set_unlimited`
    
    :Returns:
    
        `bool`
            The existing unlimited status. True and False signify
            "unlimited" and "not unlimited" repectively.
    
    **Examples:**
    
    >>> da = f.domain_axis('domainaxis1')
    >>> da.nc_is_unlimited()
    False
    >>> da.nc_set_unlimited(True)
    >>> da.nc_is_unlimited()
    True
    >>> da.nc_set_unlimited(False)
    False
    >>> da.nc_is_unlimited()
    True

        '''
        return self._get_component('netcdf').get('unlimited', False)


    def nc_set_unlimited(self, value):
        '''Set the unlimited status of the a netCDF dimension.
    
    By default output netCDF dimensions are not unlimited. The status
    is used by the `write` function.
    
    .. versionadded:: 1.7.4
    
    .. seealso:: `nc_is_unlimited`
    
    :Parameters:
    
        value: `bool`
            The new unlimited status. True and False signify
            "unlimited" and "not unlimited" repectively.

    :Returns:
    
        `None`
    
    **Examples:**

    >>> da = f.domain_axis('domainaxis1')
    >>> da.nc_is_unlimited()
    False
    >>> da.nc_set_unlimited(True)
    >>> da.nc_is_unlimited()
    True
    >>> da.nc_set_unlimited(False)
    False
    >>> da.nc_is_unlimited()
    True

        '''
        self._get_component('netcdf')['unlimited'] = bool(value)


#--- End: class




from builtins import super
from future.utils import with_metaclass

import abc

from copy import deepcopy

from . import Parameters


class ParametersDomainAncillaries(with_metaclass(abc.ABCMeta, Parameters)):
    '''Abstract base class for a collection of named parameters and named
domain ancillary constructs.

.. versionadded:: 1.7.0

    '''

    def __init__(self, parameters=None, domain_ancillaries=None,
                 source=None, copy=True):
        '''**Initialization**

:Parameters:

    parameters: `dict`, optional
       Set parameters. The dictionary keys are term names, with
       corresponding parameter values. Ignored if the *source*
       parameter is set.

       *Parameter example:*
         ``parameters={'earth_radius': 6371007.}``

       Parameters may also be set after initialisation with the
       `set_parameters` and `set_parameter` methods.

    domain_ancillaries: `dict`, optional
       Set references to domain ancillary constructs. The dictionary
       keys are term names, with corresponding domain ancillary
       construct keys. Ignored if the *source* parameter is set.

       *Parameter example:*
         ``domain_ancillaries={'orog': 'domainancillary2'}``

       Domain ancillaries may also be set after initialisation with
       the `set_domain_ancillaries` and `set_domain_ancillary`
       methods.

    source: optional
        Initialize the parameters and domain ancillary terms from
        those of *source*.

    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__(parameters=parameters, source=source,
                         copy=copy)

        self._set_component('domain_ancillaries', {}, copy=False)

        if source:
            try:
                domain_ancillaries = source.domain_ancillaries()
            except AttributeError:
                domain_ancillaries = None
        #--- End: if
        
        if domain_ancillaries is None:
            domain_ancillaries = {}
        elif copy:
            domain_ancillaries = domain_ancillaries.copy()
            for key, value in list(domain_ancillaries.items()):
                domain_ancillaries[key] = deepcopy(value)
        #--- End: if
            
        self.set_domain_ancillaries(domain_ancillaries, copy=False)
    #--- End: def

    def clear_domain_ancillaries(self):
        '''Remove all domain_ancillaries.

.. versionadded:: 1.7.0

.. seealso:: `del_domain_ancillary`, `domain_ancillaries`,
             `set_domain_ancillaries`

:Returns:

    `dict`
        The domain ancillaries that have been removed.

**Examples:**

>>> old = f.clear_domain_ancillaries()
>>> old
{'a': 'domainancillary0',
 'b': 'domainancillary1',
 'orog': 'domainancillary2'}
>>> f.set_domain_ancillaries(old)
>>> f.domain_ancillaries()
{'a': 'domainancillary0',
 'b': 'domainancillary1',
 'orog': 'domainancillary2'}

        '''
        out = self._get_component('domain_ancillaries')
        self._set_component('domain_ancillaries', {})
        return out.copy()
    #--- End: def
    
#    def copy(self):
#        '''Return a deep copy.
#
#``f.copy()`` is equivalent to ``copy.deepcopy(f)``.
#
#.. versionadded:: 1.7.0
#
#:Returns:
#
#        The deep copy.
#
#**Examples:**
#
#>>> g = f.copy()
#
#        '''
#        return type(self)(source=self, copy=True) #, _use_data=data)
#    #--- End: def

    def del_domain_ancillary(self, domain_ancillary,
                             default=ValueError()):
        '''Delete a domain ancillary.

.. versionadded:: 1.7.0

.. seealso:: `domain_ancillaries`, `get_domain_ancillary`,
             `set_domain_ancillary`

:Parameters:

    domain_ancillary: `str`
        The name of the domain ancillary to be deleted.

        *Parameter example:*
           ``domain_ancillary='orog'``

    default: optional
        Return the value of the *default* parameter if the domain
        ancillary term has not been set. If set to an `Exception`
        instance then it will be raised instead.

:Returns:

        The removed domain ancillary key.

**Examples:**

>>> c.set_domain_ancillary('orog', 'domainancillary2')
>>> c.has_domain_ancillary('orog')
True
>>> c.get_domain_ancillary('orog')
domainancillary2'   
>>> c.del_domain_ancillary('orog')
domainancillary2'   
>>> c.has_domain_ancillaryr('orog')
False
>>> print(c.del_domain_ancillaryy('orog', None))
None
>>> print(c.get_domain_ancillary('orog', None))
None

        '''
        try:
            return self._get_component('domain_ancillaries').pop(
                domain_ancillary)
        except KeyError:
            return self._default(default,
                                 "{!r} has no {!r} domain ancillary".format(
                                     self.__class__.__name__, domain_ancillary))
    #--- End: def

    def domain_ancillaries(self):
        '''Return all domain_ancillaries.

.. versionadded:: 1.7.0

.. seealso:: `clear_domain_ancillaries`, `get_domain_ancillary`,
             `has_domain_ancillaryr` `set_domain_ancillaries`

:Returns:

    `dict`
        The domain ancillaries.

**Examples:**

>>> old = f.clear_domain_ancillaries()
>>> old
{'a': 'domainancillary0',
 'b': 'domainancillary1',
 'orog': 'domainancillary2'}
>>> f.set_domain_ancillaries(old)
>>> f.domain_ancillaries()
{'a': 'domainancillary0',
 'b': 'domainancillary1',
 'orog': 'domainancillary2'}

        '''
        return self._get_component('domain_ancillaries').copy()
    #--- End: def

    def get_domain_ancillary(self, domain_ancillary,
                             default=ValueError()):
        '''Return a domain ancillary term.

.. versionadded:: 1.7.0

.. seealso:: `del_domain_ancillary`, `domain_ancillaries`,
             `set_domain_ancillary`

:Parameters:

    domain_ancillary: `str`
        The name of the term.

    default: optional
        Return the value of the *default* parameter if the domain
        ancillary term has not been set. If set to an `Exception`
        instance then it will be raised instead.

:Returns:

        The domain ancillary construct key.

**Examples:**

>>> c.set_domain_ancillary('orog', 'domainancillary2')
>>> c.has_domain_ancillary('orog')
True
>>> c.get_domain_ancillary('orog')
domainancillary2'   
>>> c.del_domain_ancillary('orog')
domainancillary2'   
>>> c.has_domain_ancillaryr('orog')
False
>>> print(c.del_domain_ancillaryy('orog', None))
None
>>> print(c.get_domain_ancillary('orog', None))
None

        '''        
        try:
            return self._get_component('domain_ancillaries')[domain_ancillary]
        except KeyError:
            return self._default(default,
                                 "{!r} has no {!r} domain ancillary".format(
                                     self.__class__.__name__, domain_ancillary))
    #--- End: def
    
    def has_domain_ancillary(self, domain_ancillary):
        '''Whether a domain ancillary has been set.

.. versionadded:: 1.7.0

.. seealso:: `del_domain_ancillary`, `domain_ancillaries`,
             `has_domain_ancillary`, `set_domain_ancillary`

:Parameters:

    domain_ancillary: `str`
        The name of the term.

    default: optional
        Return the value of the *default* parameter if the domain
        ancillary term has not been set. If set to an `Exception`
        instance then it will be raised instead.

:Returns:

        The domain ancillary construct key.

**Examples:**

>>> c.set_domain_ancillary('orog', 'domainancillary2')
>>> c.has_domain_ancillary('orog')
True
>>> c.get_domain_ancillary('orog')
domainancillary2'   
>>> c.del_domain_ancillary('orog')
domainancillary2'   
>>> c.has_domain_ancillaryr('orog')
False
>>> print(c.del_domain_ancillaryy('orog', None))
None
>>> print(c.get_domain_ancillary('orog', None))
None

        '''
        return domain_ancillary in self._get_component('domain_ancillaries')
    #--- End: def
    
    def set_domain_ancillaries(self, domain_ancillaries, copy=True):
        '''Set domain_ancillaries.

.. versionadded:: 1.7.0

.. seealso:: `clear_domain_ancillaries`, `domain_ancillaries`,
             `set_domain_ancillary`

:Parameters:

    domain_ancillaries: `dict` 
        Store the domain ancillaries from the dictionary supplied.

        *Parameter example:*
          ``domain_ancillaries={'earth_radius': 6371007}``

    copy: `bool`, optional
        If False then any parameter values provided by the
        *domain_ancillaries* parameter are not copied before
        insertion. By default they are deep copied.

:Returns:

    `None`

**Examples:**

>>> old = f.clear_domain_ancillaries()
>>> old
{'a': 'domainancillary0',
 'b': 'domainancillary1',
 'orog': 'domainancillary2'}
>>> f.set_domain_ancillaries(old)
>>> f.domain_ancillaries()
{'a': 'domainancillary0',
 'b': 'domainancillary1',
 'orog': 'domainancillary2'}

        '''
        if copy:
            domain_ancillaries = deepcopy(domain_ancillaries)                
        else:
            domain_ancillaries = domain_ancillaries.copy()
        
        self._get_component('domain_ancillaries').update(domain_ancillaries)
    #--- End: def

    def set_domain_ancillary(self, term, value, copy=True):
        '''Set an domain ancillary-valued term.

.. versionadded:: 1.7.0

.. seealso:: `del_domain_ancillary`, `domain_ancillaries`,
             `get_domain_ancillary`

:Parameters:

    term: `str`
        The name of the term to be set.

    value:
        The value for the term.

    copy: `bool`, optional
        If True then set a deep copy of *value*.

:Returns:

    `None`

**Examples:**

>>> c.set_domain_ancillary('orog', 'domainancillary2')
>>> c.has_domain_ancillary('orog')
True
>>> c.get_domain_ancillary('orog')
domainancillary2'   
>>> c.del_domain_ancillary('orog')
domainancillary2'   
>>> c.has_domain_ancillaryr('orog')
False
>>> print(c.del_domain_ancillaryy('orog', None))
None
>>> print(c.get_domain_ancillary('orog', None))
None

        '''
        if copy:
            value = deepcopy(value)
            
        self._get_component('domain_ancillaries')[term] = value
    #--- End: def
        
#--- End: class



from builtins import super

from . import mixin
from . import core


class InteriorRing(mixin.NetCDFDimension,
                   mixin.NetCDFVariable,
                   mixin.PropertiesData,
                   core.InteriorRing):
    '''An interior ring array with properties.

If a cell is composed of multiple polygon parts, an individual polygon
may define an "interior ring", i.e. a region that is to be omitted
from, as opposed to included in, the cell extent. In this case an
interior ring array is required that records whether each polygon is
to be included or excluded from the cell, and is supplied by an
interior ring variable in CF-netCDF. The interior ring array spans the
same domain axis constructs as its coordinate array, with the addition
of an extra dimension that indexes the parts for each cell. For
example, a cell describing the land area surrounding a lake would
require two polygon parts: one defines the outer boundary of the land
area; the other, recorded as an interior ring, is the lake boundary,
defining the inner boundary of the land area.

**NetCDF interface**

The netCDF variable name of the interior ring variable may be accessed
with the `nc_set_variable`, `nc_get_variable`, `nc_del_variable` and
`nc_has_variable` methods.

The name of the netCDF dimension spanned by the interior ring
variable's data (which does not correspond to a domain axis construct)
may be accessed with the `nc_set_dimension`, `nc_get_dimension`,
`nc_del_dimension` and `nc_has_dimension` methods.

.. versionadded:: 1.8.0

    '''
    def __init__(self, properties=None, data=None, source=None,
                 copy=True, _use_data=True):
        '''**Initialization**

    properties: `dict`, optional
        Set descriptive properties. The dictionary keys are property
        names, with corresponding values. Ignored if the *source*
        parameter is set.

        *Parameter example:*
          ``properties={'long_name': 'which station this obs is for'}``

        Properties may also be set after initialisation with the
        `set_properties` and `set_property` methods.

    data: `Data`, optional
        Set the data array. Ignored if the *source* parameter is set.

        The data array may also be set after initialisation with the
        `set_data` method.

    source: optional
        Initialize the properties and data from those of *source*.

    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__(properties=properties, data=data,
                         source=source, copy=copy,
                         _use_data=_use_data)
        
        self._initialise_netcdf(source)
    #--- End: def
    
    def dump(self, display=True, _key=None, _title=None,
             _create_title=True, _prefix='', _level=0,
             _omit_properties=None, _axes=None, _axis_names=None):
        '''A full description of the interior ring variable.

Returns a description of all properties, including those of
components, and provides selected values of all data arrays.

.. versionadded:: 1.8.0

:Parameters:

    display: `bool`, optional
        If False then return the description as a string. By default
        the description is printed.

:Returns:

    `None` or `str`
        The description. If *display* is True then the description is
        printed and `None` is returned. Otherwise the description is
        returned as a string.

        '''
        if _create_title and _title is None: 
            _title = 'Interior Ring: ' + self.identity(default='')

        return super().dump(display=display, _key=_key,
                            _omit_properties=_omit_properties,
                            _prefix=_prefix, _level=_level,
                            _title=_title,
                            _create_title=_create_title, _axes=_axes,
                            _axis_names=_axis_names)
    #--- End: def
    
#--- End: class



from builtins import super

from . import mixin
from . import core


class Index(mixin.NetCDFVariable,
            mixin.NetCDFDimension,
            mixin.NetCDFSampleDimension,
            mixin.PropertiesData,
            core.abstract.PropertiesData):
    '''An index variable required to uncompress a ragged array.

A collection of features stored using an indexed ragged array combines
all features along a single dimension (the sample dimension) such that
the values of each feature in the collection are interleaved.

The information needed to uncompress the data is stored in an index
variable that specifies the feature that each element of the sample
dimension belongs to.

**NetCDF interface**

The netCDF variable name of the index variable may be accessed with
the `nc_set_variable`, `nc_get_variable`, `nc_del_variable` and
`nc_has_variable` methods.

The name of the netCDF dimension spanned by the index variable's data
(which does not correspond to a domain axis construct) may be accessed
with the `nc_set_dimension`, `nc_get_dimension`, `nc_del_dimension`
and `nc_has_dimension` methods.

The name of the netCDF sample dimension spanned by the compressed data
(which does not correspond to a domain axis contract) may be accessed
with the `nc_set_sample_dimension`, `nc_get_sample_dimension`,
`nc_del_sample_dimension` and `nc_has_sample_dimension` methods.

   .. note:: The netCDF sample dimension and the netCDF dimension
             spanned by the index variable's data are should be the
             same, unless the compressed data is an indexed contiguous
             ragged array, in which case they must be different.

The name of the netCDF instance dimension (that is stored in the
"instance_dimension" netCDF attribute) is accessed via the
corresponding domain axis construct.

.. versionadded:: 1.7.0

    '''
    def __init__(self, properties=None, data=None, source=None,
                 copy=True, _use_data=True):
        '''**Initialization**

:Parameters:

    properties: `dict`, optional
        Set descriptive properties. The dictionary keys are property
        names, with corresponding values. Ignored if the *source*
        parameter is set.

        *Parameter example:*
          ``properties={'long_name': 'which station this obs is for'}``

        Properties may also be set after initialisation with the
        `set_properties` and `set_property` methods.

    data: `Data`, optional
        Set the data array. Ignored if the *source* parameter is set.

        The data array may also be set after initialisation with the
        `set_data` method.

    source: optional
        Initialize the properties and data from those of *source*.

    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__(properties=properties, data=data,
                         source=source, copy=copy,
                         _use_data=_use_data)
        
        self._initialise_netcdf(source)
    #--- End: def
    
    def dump(self, display=True, _key=None, _title=None,
             _create_title=True, _prefix='', _level=0,
             _omit_properties=None, _axes=None, _axis_names=None):
        '''A full description of the index variable.

Returns a description of all properties, including those of
components, and provides selected values of all data arrays.

.. versionadded:: 1.7.0

:Parameters:

    display: `bool`, optional
        If False then return the description as a string. By default
        the description is printed.

:Returns:

    `None` or `str`
        The description. If *display* is True then the description is
        printed and `None` is returned. Otherwise the description is
        returned as a string.

        '''
        if _create_title and _title is None: 
            _title = 'Index: ' + self.identity(default='')

        return super().dump(display=display, _key=_key,
                            _omit_properties=_omit_properties,
                            _prefix=_prefix, _level=_level,
                            _title=_title,
                            _create_title=_create_title, _axes=_axes,
                            _axis_names=_axis_names)
    #--- End: def
    
#--- End: class



from __future__ import print_function
import datetime
import inspect
import os
import tempfile
import unittest

from distutils.version import LooseVersion

import numpy
import netCDF4

import cfdm

VN = cfdm.CF()
if LooseVersion(VN) < LooseVersion('1.8'):
    VN = '1.8'

def _make_geometry_1_file(filename):
    '''See n.comment for details.

    '''
    n = netCDF4.Dataset(filename, 'w', format='NETCDF3_CLASSIC')
    
    n.Conventions = 'CF-'+VN
    n.featureType = 'timeSeries'
    n.comment     = "Make a netCDF file with 2 node coordinates variables, each of which has a corresponding auxiliary coordinate variable."
    
    time     = n.createDimension('time'    , 4)
    instance = n.createDimension('instance', 2)
    node     = n.createDimension('node'    , 5)
    
    t =  n.createVariable('time', 'i4', ('time',))
    t.units = "seconds since 2016-11-07 20:00 UTC" 
    t[...] = [1, 2, 3, 4 ]

    lat = n.createVariable('lat', 'f8', ('instance',))
    lat.standard_name = "latitude"
    lat.units = "degrees_north"
    lat.nodes = "y"
    lat[...] = [30, 50]

    lon = n.createVariable('lon', 'f8', ('instance',))
    lon.standard_name = "longitude"
    lon.units = "degrees_east"
    lon.nodes = "x"
    lon[...] = [10, 60]    

    datum = n.createVariable('datum', 'i4', ())
    datum.grid_mapping_name = "latitude_longitude"
    datum.longitude_of_prime_meridian = 0.0
    datum.semi_major_axis = 6378137.0
    datum.inverse_flattening = 298.257223563
    
    geometry_container = n.createVariable('geometry_container', 'i4', ());
    geometry_container.geometry_type = "line"
    geometry_container.node_count = "node_count"
    geometry_container.node_coordinates = "x y"
    geometry_container.geometry_dimension = "instance"
    
    node_count = n.createVariable('node_count', 'i4', ('instance',))
    node_count[...] = [3, 2]
    
    x = n.createVariable('x', 'f8', ('node',))
    x.units = "degrees_east"
    x.standard_name = "longitude"
    x.axis = "X"
    x[...] = [30, 10, 40, 50, 50]
    
    y = n.createVariable('y', 'f8', ('node',))
    y.units = "degrees_north"
    y.standard_name = "latitude"
    y.axis = "Y"
    y[...] = [10, 30, 40, 60, 50]
    
    pr = n.createVariable('pr', 'f8', ('instance', 'time'))
    pr.standard_name = 'precipitation_amount'
    pr.units = 'kg m-2'
    pr.coordinates = "time lat lon"
    pr.grid_mapping = "datum"
    pr.geometry = "geometry_container"
    pr[...] = [[1, 2, 3, 4],
                       [5, 6, 7, 8]]

    someData_2 = n.createVariable('someData_2', 'f8', ('instance', 'time'))
    someData_2.coordinates = "time lat lon"
    someData_2.grid_mapping = "datum"
    someData_2.geometry = "geometry_container"
    someData_2[...] = [[10, 20, 30, 40],
                       [50, 60, 70, 80]]

    n.close()
    
    return filename
#--- End: def

def _make_geometry_2_file(filename):        
    '''See n.comment for details

    '''
    n = netCDF4.Dataset(filename, 'w', format='NETCDF3_CLASSIC')
    
    n.Conventions = 'CF-'+VN
    n.featureType = 'timeSeries'
    n.comment     = 'A netCDF file with 3 node coordinates variables, only two of which have a corresponding auxiliary coordinate variable.'
   
    time     = n.createDimension('time'    , 4)
    instance = n.createDimension('instance', 2)
    node     = n.createDimension('node'    , 5)
    
    t =  n.createVariable('time', 'i4', ('time',))
    t.units = "seconds since 2016-11-07 20:00 UTC" 
    t[...] = [1, 2, 3, 4 ]

    lat = n.createVariable('lat', 'f8', ('instance',))
    lat.standard_name = "latitude"
    lat.units = "degrees_north"
    lat.nodes = "y"
    lat[...] = [30, 50]

    lon = n.createVariable('lon', 'f8', ('instance',))
    lon.standard_name = "longitude"
    lon.units = "degrees_east"
    lon.nodes = "x"
    lon[...] = [10, 60]    

    datum = n.createVariable('datum', 'i4', ())
    datum.grid_mapping_name = "latitude_longitude"
    datum.longitude_of_prime_meridian = 0.0
    datum.semi_major_axis = 6378137.0
    datum.inverse_flattening = 298.257223563
    
    geometry_container = n.createVariable('geometry_container', 'i4', ());
    geometry_container.geometry_type = "line"
    geometry_container.node_count = "node_count"
    geometry_container.node_coordinates = "x y z"
    geometry_container.geometry_dimension = "instance"
    
    node_count = n.createVariable('node_count', 'i4', ('instance',))
    node_count[...] = [3, 2]
    
    x = n.createVariable('x', 'f8', ('node',))
    x.units = "degrees_east"
    x.standard_name = "longitude"
    x.axis = "X"
    x[...] = [30, 10, 40, 50, 50]
    
    y = n.createVariable('y', 'f8', ('node',))
    y.units = "degrees_north"
    y.standard_name = "latitude"
    y.axis = "Y"
    y[...] = [10, 30, 40, 60, 50]
    
    z = n.createVariable('z', 'f8', ('node',))
    z.units = "m"
    z.standard_name = "altitude"
    z.axis = "Z"
    z[...] = [100, 150, 200, 125, 80]
    
    someData = n.createVariable('someData', 'f8', ('instance', 'time'))
    someData.coordinates = "time lat lon"
    someData.grid_mapping = "datum"
    someData.geometry = "geometry_container"
    someData[...] = [[1, 2, 3, 4],
                     [5, 6, 7, 8]]

    someData_2 = n.createVariable('someData_2', 'f8', ('instance', 'time'))
    someData_2.coordinates = "time lat lon"
    someData_2.grid_mapping = "datum"
    someData_2.geometry = "geometry_container"
    someData_2[...] = [[1, 2, 3, 4],
                       [5, 6, 7, 8]]

    n.close()
    
    return filename
#--- End: def

def _make_geometry_3_file(filename):        
    '''See n.comment for details

    '''
    n = netCDF4.Dataset(filename, 'w', format='NETCDF3_CLASSIC')
    
    n.Conventions = 'CF-'+VN
    n.featureType = 'timeSeries'
    n.comment     = "A netCDF file with 3 node coordinates variables, each of which contains only one point, only two of which have a corresponding auxiliary coordinate variables. There is no node count variable."

   
    time     = n.createDimension('time'    , 4)
    instance = n.createDimension('instance', 3)
    node     = n.createDimension('node'    , 3)
    
    t =  n.createVariable('time', 'i4', ('time',))
    t.units = "seconds since 2016-11-07 20:00 UTC" 
    t[...] = [1, 2, 3, 4 ]

    lat = n.createVariable('lat', 'f8', ('instance',))
    lat.standard_name = "latitude"
    lat.units = "degrees_north"
    lat.nodes = "y"
    lat[...] = [30, 50, 70]
    
    lon = n.createVariable('lon', 'f8', ('instance',))
    lon.standard_name = "longitude"
    lon.units = "degrees_east"
    lon.nodes = "x"
    lon[...] = [10, 60, 80]    
    
    datum = n.createVariable('datum', 'i4', ())
    datum.grid_mapping_name = "latitude_longitude"
    datum.longitude_of_prime_meridian = 0.0
    datum.semi_major_axis = 6378137.0
    datum.inverse_flattening = 298.257223563
    
    geometry_container = n.createVariable('geometry_container', 'i4', ());
    geometry_container.geometry_type = "point"
    geometry_container.node_coordinates = "x y z"
    geometry_container.geometry_dimension = "instance"
    
    x = n.createVariable('x', 'f8', ('node',))
    x.units = "degrees_east"
    x.standard_name = "longitude"
    x.axis = "X"
    x[...] = [30, 10, 40]
    
    y = n.createVariable('y', 'f8', ('node',))
    y.units = "degrees_north"
    y.standard_name = "latitude"
    y.axis = "Y"
    y[...] = [10, 30, 40]
    
    z = n.createVariable('z', 'f8', ('node',))
    z.units = "m"
    z.standard_name = "altitude"
    z.axis = "Z"
    z[...] = [100, 150, 200]
    
    someData_1 = n.createVariable('someData_1', 'f8', ('instance', 'time'))
    someData_1.coordinates = "lat lon"
    someData_1.grid_mapping = "datum"
    someData_1.geometry = "geometry_container"
    someData_1[...] = [[1,  2,  3,  4],
                       [5,  6,  7,  8],
                       [9, 10, 11, 12]]
    
    someData_2 = n.createVariable('someData_2', 'f8', ('instance', 'time'))
    someData_2.coordinates = "lat lon"
    someData_2.grid_mapping = "datum"
    someData_2.geometry = "geometry_container"
    someData_2[...] = [[10,  20,  30,  40],
                       [50,  60,  70,  80],
                       [90, 100, 110, 120]]
    
    n.close()
    
    return filename
#--- End: def

def _make_geometry_4_file(filename):
    '''See n.comment for details.
    '''
    n = netCDF4.Dataset(filename, 'w', format='NETCDF3_CLASSIC')
    
    n.Conventions = 'CF-'+VN
    n.featureType = 'timeSeries'
    n.comment     = "A netCDF file with 2 node coordinates variables, none of which have a corresponding auxiliary coordinate variable."
    
    time     = n.createDimension('time'    , 4)
    instance = n.createDimension('instance', 2)
    node     = n.createDimension('node'    , 5)
    strlen   = n.createDimension('strlen'  , 2)

    # Variables
    t = n.createVariable('time', 'i4', ('time',))
    t.standard_name = "time" 
    t.units = "days since 2000-01-01"
    t[...] = [1, 2, 3, 4]

    instance_id = n.createVariable('instance_id', 'S1', ('instance', 'strlen'))
    instance_id.cf_role = "timeseries_id"
    instance_id[...] = [['x', '1'],
                        ['y', '2']]

    datum = n.createVariable('datum', 'i4', ())
    datum.grid_mapping_name = "latitude_longitude"
    datum.longitude_of_prime_meridian = 0.0
    datum.semi_major_axis = 6378137.0
    datum.inverse_flattening = 298.257223563
    
    geometry_container = n.createVariable('geometry_container', 'i4', ());
    geometry_container.geometry_type = "line"
    geometry_container.node_count = "node_count"
    geometry_container.node_coordinates = "x y"
    geometry_container.geometry_dimension = "instance"
    
    node_count = n.createVariable('node_count', 'i4', ('instance',))
    node_count[...] = [3, 2]
    
    x = n.createVariable('x', 'f8', ('node',))
    x.units = "degrees_east"
    x.standard_name = "longitude"
    x.axis = "X"
    x[...] = [30, 10, 40, 50, 50]
    
    y = n.createVariable('y', 'f8', ('node',))
    y.units = "degrees_north"
    y.standard_name = "latitude"
    y.axis = "Y"
    y[...] = [10, 30, 40, 60, 50]
    
    someData_1 = n.createVariable('someData_1', 'f8', ('instance', 'time'))
    someData_1.coordinates = "instance_id"
    someData_1.grid_mapping = "datum"
    someData_1.geometry = "geometry_container"
    someData_1[...] = [[1, 2, 3, 4],
                       [5, 6, 7, 8]]

    someData_2 = n.createVariable('someData_2', 'f8', ('instance', 'time'))
    someData_2.coordinates = "instance_id"
    someData_2.grid_mapping = "datum"
    someData_2.geometry = "geometry_container"
    someData_2[...] = [[10, 20, 30, 40],
                       [50, 60, 70, 80]]

    n.close()
    
    return filename
#--- End: def

#def _make_interior_ring_file_DEL(filename):        
#    '''See n.comment for details.
#    '''
#    n = netCDF4.Dataset(filename, 'w', format='NETCDF3_CLASSIC')
#    
#    # Global arttributes
#    n.Conventions = 'CF-'+VN
#
#    # Dimensions
#    time     = n.createDimension('time', 4)
#    instance = n.createDimension('instance', 2)
#    node     = n.createDimension('node', 13)
#    part     = n.createDimension('part', 4)
#    strlen   = n.createDimension('strlen', 2)
#
#    # Variables
#    t = n.createVariable('time', 'i4', ('time',))
#    t.standard_name = "time" 
#    t.units = "days since 2000-01-01"
#    t[...] = [1, 2, 3, 4]
#
#    instance_id = n.createVariable('instance_id', 'S1', ('instance', 'strlen'))
#    instance_id.cf_role = "timeseries_id"
#    instance_id[...] = [['x', '1'],
#                        ['y', '2']]
#    
#    x = n.createVariable('x', 'f8', ('node',))
#    x.units = "degrees_east"
#    x.standard_name = "longitude"
#    x.axis = "X"
#    x[...] = [20, 10, 0,
#              5, 10, 15, 10,
#              20, 10, 0,
#
#              50, 40, 30]
# 
#    y = n.createVariable('y', 'f8', ('node',))
#    y.units = "degrees_north"
#    y.standard_name = "latitude"
#    y.axis = "Y"
#    y[...] = [0, 15, 0,
#              5, 10, 5, 5,
#              20, 35, 20,
#
#              0, 15, 0]
# 
#    z = n.createVariable('z', 'f8', ('node',))
#    z.units = "m"
#    z.standard_name = "altitude"
#    z.axis = "Z"
#    z[...] = [1, 2, 4,
#              2, 3, 4, 5,
#              5, 1, 4,
#
#              3, 2, 1]
# 
#    lat = n.createVariable('lat', 'f8', ('instance',))
#    lat.units = "degrees_north" 
#    lat.standard_name = "latitude"
#    lat.nodes = "y"
#    lat[...] = [25, 7]
#
#    lon = n.createVariable('lon', 'f8', ('instance',))
#    lon.units = "degrees_east"
#    lon.standard_name = "longitude"
#    lon.nodes = "x"
#    lon[...] = [10, 40]
#
#    geometry_container = n.createVariable('geometry_container', 'i4', ())
#    geometry_container.geometry_type = "polygon"
#    geometry_container.node_count = "node_count"
#    geometry_container.node_coordinates = "x y z"
#    geometry_container.grid_mapping = "datum"
#    geometry_container.coordinates = "lat lon"
#    geometry_container.part_node_count = "part_node_count"
#    geometry_container.interior_ring = "interior_ring"
#    geometry_container.geometry_dimension = "instance"
#    
#    node_count = n.createVariable('node_count', 'i4', ('instance'))
#    node_count[...] = [10, 3]
#
#    part_node_count = n.createVariable('part_node_count', 'i4', ('part'))
#    part_node_count[...] = [3, 4, 3,
#                            3]
#    
#    interior_ring = n.createVariable('interior_ring', 'i4', ('part'))
#    interior_ring[...] = [0, 1, 0, 0]
#
#    datum = n.createVariable('datum', 'f4', ())
#    datum.grid_mapping_name = "latitude_longitude"
#    datum.semi_major_axis = 6378137.
#    datum.inverse_flattening = 298.257223563
#    datum.longitude_of_prime_meridian = 0.
#    
#    pr = n.createVariable('pr', 'f8', ('instance', 'time'))
#    pr.standard_name = "preciptitation_amount"
#    pr.standard_units = "kg m-2"
#    pr.coordinates = "time lat lon instance_id"
#    pr.grid_mapping = "datum"
#    pr.geometry = "geometry_container"
#    pr[...]= [[1, 2, 3, 4],
#              [5, 6, 7, 8]]
#  
#    n.close()
#    
#    return filename
##--- End: def

def _make_interior_ring_file(filename):        
    '''See n.comment for details.
    '''
    n = netCDF4.Dataset(filename, 'w', format='NETCDF3_CLASSIC')
    
    # Global arttributes
    n.Conventions = 'CF-'+VN
    n.featureType = 'timeSeries'
    n.comment = 'TODO'

    # Dimensions
    time     = n.createDimension('time', 4)
    instance = n.createDimension('instance', 2)
    node     = n.createDimension('node', 13)
    part     = n.createDimension('part', 4)
    strlen   = n.createDimension('strlen', 2)

    # Variables
    t = n.createVariable('time', 'i4', ('time',))
    t.standard_name = "time" 
    t.units = "days since 2000-01-01"
    t[...] = [1, 2, 3, 4]

    instance_id = n.createVariable('instance_id', 'S1', ('instance', 'strlen'))
    instance_id.cf_role = "timeseries_id"
    instance_id[...] = [['x', '1'],
                        ['y', '2']]
    
    x = n.createVariable('x', 'f8', ('node',))
    x.units = "degrees_east"
    x.standard_name = "longitude"
    x.axis = "X"
    x[...] = [20, 10, 0,
              5, 10, 15, 10,
              20, 10, 0,

              50, 40, 30]
 
    y = n.createVariable('y', 'f8', ('node',))
    y.units = "degrees_north"
    y.standard_name = "latitude"
    y.axis = "Y"
    y[...] = [0, 15, 0,
              5, 10, 5, 5,
              20, 35, 20,

              0, 15, 0]
 
    z = n.createVariable('z', 'f8', ('node',))
    z.units = "m"
    z.standard_name = "altitude"
    z.axis = "Z"
    z[...] = [1, 2, 4,
              2, 3, 4, 5,
              5, 1, 4,

              3, 2, 1]
 
    lat = n.createVariable('lat', 'f8', ('instance',))
    lat.units = "degrees_north" 
    lat.standard_name = "latitude"
    lat.nodes = "y"
    lat[...] = [25, 7]

    lon = n.createVariable('lon', 'f8', ('instance',))
    lon.units = "degrees_east"
    lon.standard_name = "longitude"
    lon.nodes = "x"
    lon[...] = [10, 40]

    geometry_container = n.createVariable('geometry_container', 'i4', ())
    geometry_container.geometry_type = "polygon"
    geometry_container.node_count = "node_count"
    geometry_container.node_coordinates = "x y z"
    geometry_container.grid_mapping = "datum"
    geometry_container.coordinates = "lat lon"
    geometry_container.part_node_count = "part_node_count"
    geometry_container.interior_ring = "interior_ring"
    geometry_container.geometry_dimension = "instance"
    
    node_count = n.createVariable('node_count', 'i4', ('instance'))
    node_count[...] = [10, 3]

    part_node_count = n.createVariable('part_node_count', 'i4', ('part'))
    part_node_count[...] = [3, 4, 3,
                            3]
    
    interior_ring = n.createVariable('interior_ring', 'i4', ('part'))
    interior_ring[...] = [0, 1, 0, 0]

    datum = n.createVariable('datum', 'f4', ())
    datum.grid_mapping_name = "latitude_longitude"
    datum.semi_major_axis = 6378137.
    datum.inverse_flattening = 298.257223563
    datum.longitude_of_prime_meridian = 0.
    
    pr = n.createVariable('pr', 'f8', ('instance', 'time'))
    pr.standard_name = "preciptitation_amount"
    pr.standard_units = "kg m-2"
    pr.coordinates = "time lat lon instance_id"
    pr.grid_mapping = "datum"
    pr.geometry = "geometry_container"
    pr[...]= [[1, 2, 3, 4],
                    [5, 6, 7, 8]]
  
    someData_2 = n.createVariable('someData_2', 'f8', ('instance', 'time'))
    someData_2.coordinates = "time lat lon instance_id"
    someData_2.grid_mapping = "datum"
    someData_2.geometry = "geometry_container"
    someData_2[...]= [[1, 2, 3, 4],
                      [5, 6, 7, 8]]
  
    n.close()
    
    return filename
#--- End: def

geometry_1_file    = _make_geometry_1_file('geometry_1.nc')
geometry_2_file    = _make_geometry_2_file('geometry_2.nc')
geometry_3_file    = _make_geometry_3_file('geometry_3.nc')
geometry_4_file    = _make_geometry_4_file('geometry_4.nc')
interior_ring_file = _make_interior_ring_file('geometry_interior_ring.nc')
#interior_ring_file_DEL = _make_interior_ring_file_DEL('geometry.nc')

class DSGTest(unittest.TestCase):
    def setUp(self):
        self.geometry_1_file = geometry_1_file
        self.geometry_2_file = geometry_2_file
        self.geometry_3_file = geometry_3_file
        self.geometry_4_file = geometry_4_file
        self.geometry_interior_ring_file = interior_ring_file

        (fd, self.tempfilename) = tempfile.mkstemp(suffix='.nc', prefix='cfdm_', dir='.')
        os.close(fd)
#        self.tempfilename = 'delme.nc'

       
        self.test_only = []
#        self.test_only = ['test_node_count']
#        self.test_only = ['test_geometry_interior_ring']

 
    def tearDown(self):
        os.remove(self.tempfilename)

        
    @unittest.skipIf(cfdm.__version__ < '1.8',
                     "not supported in this library version")
    def test_node_count(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return
                
        f = cfdm.read(self.geometry_1_file, verbose=False)

        self.assertTrue(len(f) == 2, 'f = '+repr(f))
        for g in f:
            self.assertTrue(g.equals(g.copy(), verbose=True))
            self.assertTrue(len(g.auxiliary_coordinates) == 2)
        
        g = f[0]
        for axis in ('X', 'Y'):
            coord = g.construct('axis='+axis)
            self.assertTrue( coord.has_node_count(), 'axis='+axis)
            self.assertFalse(coord.has_part_node_count(), 'axis='+axis)
            self.assertFalse(coord.has_interior_ring(), 'axis='+axis)

        cfdm.write(f, self.tempfilename, Conventions='CF-'+VN, verbose=False)

        f2 = cfdm.read(self.tempfilename, verbose=False)
        self.assertTrue(len(f2) == 2, 'f2 = '+repr(f2))
        for a, b in zip(f, f2):
            self.assertTrue(a.equals(b, verbose=True))

        # Setting of node count properties
        coord = f[0].construct('axis=X')
        nc = coord.get_node_count()
        cfdm.write(f, self.tempfilename)
        nc.set_property('long_name', 'Node counts')
        cfdm.write(f, self.tempfilename)
        nc.nc_set_variable('new_var_name_X')
        cfdm.write(f, self.tempfilename)
        
        # Node count access
        c = g.construct('longitude').copy()            
        self.assertTrue(c.has_node_count())
        n = c.del_node_count() 
        self.assertFalse(c.has_node_count())
        self.assertTrue(c.get_node_count(None) == None)
        self.assertTrue(c.del_node_count(None) == None)
        c.set_node_count(n)
        self.assertTrue(c.has_node_count())
        self.assertTrue(c.get_node_count(None).equals(n, verbose=True))
        self.assertTrue(c.del_node_count(None).equals(n, verbose=True))
        self.assertFalse(c.has_node_count())
    #--- End: def

    @unittest.skipIf(cfdm.__version__ < '1.8',
                     "not supported in this library version")
    def test_geometry_2(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return
                
        f = cfdm.read(self.geometry_2_file, verbose=False)

        self.assertTrue(len(f) == 2, 'f = '+repr(f))

        for g in f:
            self.assertTrue(g.equals(g.copy(), verbose=True))
            self.assertTrue(len(g.auxiliary_coordinates) == 3)

        g = f[0]
        for axis in ('X', 'Y', 'Z'):
            coord = g.construct('axis='+axis)
            self.assertTrue( coord.has_node_count(), 'axis='+axis)
            self.assertFalse(coord.has_part_node_count(), 'axis='+axis)
            self.assertFalse(coord.has_interior_ring(), 'axis='+axis)

        cfdm.write(f, self.tempfilename, Conventions='CF-'+VN, verbose=False)

        f2 = cfdm.read(self.tempfilename, verbose=False)

        self.assertTrue(len(f2) == 2, 'f2 = '+repr(f2))
        
        for a, b in zip(f, f2):
            self.assertTrue(a.equals(b, verbose=True))
            
        # Setting of node count properties
        coord = f[0].construct('axis=X')
        nc = coord.get_node_count()
        cfdm.write(f, self.tempfilename)
        nc.set_property('long_name', 'Node counts')
        cfdm.write(f, self.tempfilename, verbose=False)
        nc.nc_set_variable('new_var_name')
        cfdm.write(f, self.tempfilename, verbose=False)
    #--- End: def

    @unittest.skipIf(cfdm.__version__ < '1.8',
                     "not supported in this library version")
    def test_geometry_3(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return
                
        f = cfdm.read(self.geometry_3_file, verbose=False)

        self.assertTrue(len(f) == 2, 'f = '+repr(f))

        for g in f:
            self.assertTrue(g.equals(g.copy(), verbose=True))
            self.assertTrue(len(g.auxiliary_coordinates) == 3)

        g = f[0]
        for axis in ('X', 'Y', 'Z'):
            coord = g.construct('axis='+axis)
            self.assertFalse(coord.has_node_count(), 'axis='+axis)
            self.assertFalse(coord.has_part_node_count(), 'axis='+axis)
            self.assertFalse(coord.has_interior_ring(), 'axis='+axis)

        cfdm.write(f, self.tempfilename, Conventions='CF-'+VN, verbose=False)

        f2 = cfdm.read(self.tempfilename, verbose=False)

        self.assertTrue(len(f2) == 2, 'f2 = '+repr(f2))
        
        for a, b in zip(f, f2):
            self.assertTrue(a.equals(b, verbose=True))
    #--- End: def

    @unittest.skipIf(cfdm.__version__ < '1.8',
                     "not supported in this library version")
    def test_geometry_4(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return
                
        f = cfdm.read(self.geometry_4_file, verbose=False)

        self.assertTrue(len(f) == 2, 'f = '+repr(f))

        for g in f:
            self.assertTrue(g.equals(g.copy(), verbose=True))
            self.assertTrue(len(g.auxiliary_coordinates) == 3)

        for axis in ('X', 'Y'):
            coord = g.construct('axis='+axis)
            self.assertTrue( coord.has_node_count(), 'axis='+axis)
            self.assertFalse(coord.has_part_node_count(), 'axis='+axis)
            self.assertFalse(coord.has_interior_ring(), 'axis='+axis)

        cfdm.write(f, self.tempfilename, Conventions='CF-'+VN, verbose=False)

        f2 = cfdm.read(self.tempfilename, verbose=False)

        self.assertTrue(len(f2) == 2, 'f2 = '+repr(f2))
        
        for a, b in zip(f, f2):
            self.assertTrue(a.equals(b, verbose=True))

        # Setting of node count properties
        coord = f[0].construct('axis=X')
        nc = coord.get_node_count()
        cfdm.write(f, self.tempfilename)
        nc.set_property('long_name', 'Node counts')
        cfdm.write(f, self.tempfilename, verbose=False)
        nc.nc_set_variable('new_var_name')
        cfdm.write(f, self.tempfilename, verbose=False)
    #--- End: def

    @unittest.skipIf(cfdm.__version__ < '1.8',
                     "not supported in this library version")
    def test_geometry_interior_ring(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cfdm.read(self.geometry_interior_ring_file, verbose=False)

        self.assertTrue(len(f) == 2, 'f = '+repr(f))

        for g in f:
            self.assertTrue(g.equals(g.copy(), verbose=True))
            self.assertTrue(len(g.auxiliary_coordinates) == 4)

        g = f[0]
        for axis in ('X', 'Y', 'Z'):
            coord = g.construct('axis='+axis)
            self.assertTrue(coord.has_node_count(), 'axis='+axis)
            self.assertTrue(coord.has_part_node_count(), 'axis='+axis)
            self.assertTrue(coord.has_interior_ring(), 'axis='+axis)

        cfdm.write(f, self.tempfilename, Conventions='CF-'+VN, verbose=False)

        f2 = cfdm.read(self.tempfilename, verbose=False)

        self.assertTrue(len(f2) == 2, 'f2 = '+repr(f2))
        
        for a, b in zip(f, f2):
            self.assertTrue(a.equals(b, verbose=True))

        # Interior ring component
        c = g.construct('longitude')
        
        self.assertTrue(c.interior_ring.equals(g.construct('longitude').get_interior_ring()))
        self.assertTrue(c.interior_ring.data.ndim == c.data.ndim + 1)
        self.assertTrue(c.interior_ring.data.shape[0] == c.data.shape[0])
        
        _ = g.dump(display=False)

        d = c.insert_dimension(0)
        self.assertTrue(d.data.shape == (1,) + c.data.shape)
        self.assertTrue(d.interior_ring.data.shape == (1,) + c.interior_ring.data.shape)

        e = d.squeeze(0)
        self.assertTrue(e.data.shape == c.data.shape)
        self.assertTrue(e.interior_ring.data.shape == c.interior_ring.data.shape)

        t = d.transpose()
        self.assertTrue(t.data.shape == d.data.shape[::-1], (t.data.shape, c.data.shape[::-1]))
        self.assertTrue(t.interior_ring.data.shape == d.interior_ring.data.shape[-2::-1] + (d.interior_ring.data.shape[-1],))

        # Subspacing
        g = g[1, ...]
        c = g.construct('longitude')
        self.assertTrue(c.interior_ring.data.shape[0] == 1)
        self.assertTrue(c.interior_ring.data.ndim == c.data.ndim + 1)
        self.assertTrue(c.interior_ring.data.shape[0] == c.data.shape[0])        

        # Setting of node count properties
        coord = f[0].construct('axis=Y')
        nc = coord.get_node_count()
        nc.set_property('long_name', 'Node counts')
        cfdm.write(f, self.tempfilename)
        
        nc.nc_set_variable('new_var_name')
        cfdm.write(f, self.tempfilename)

        # Setting of part node count properties
        coord = f[0].construct('axis=Z')
        pnc = coord.get_part_node_count()
        pnc.set_property('long_name', 'Part node counts')
        cfdm.write(f, self.tempfilename)
        
        pnc.nc_set_variable('new_var_name')
        cfdm.write(f, self.tempfilename)
        
        pnc.nc_set_dimension('new_dim_name')
        cfdm.write(f, self.tempfilename)
    #--- End: def    
#--- End: class


if __name__ == '__main__':
    print('Run date:', datetime.datetime.utcnow())
    cfdm.environment()
    print()
    unittest.main(verbosity=2)
    



from __future__ import print_function
import collections
import datetime
import inspect
import os
import re
import unittest

import numpy

import cfdm

class FieldTest(unittest.TestCase):
    def setUp(self):
        self.filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                     'test_file.nc')
        self.contiguous = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                       'DSG_timeSeries_contiguous.nc')
        
        f = cfdm.read(self.filename)
        self.assertTrue(len(f)==1, 'f={}'.format(f))
        self.f = f[0]

        self.test_only = []
#        self.test_only = ['test_Field_constructs']
#        self.test_only = ['test_Field_domain_axes']
#        self.test_only = ['test_Field_axes','test_Field_data_axes']
#        self.test_only = ['test_Field___getitem__']
#        self.test_only = ['test_Field___setitem__']
#        self.test_only = ['test_Field_field']


    def test_Field__repr__str__dump_construct_type(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f
        
        _ = repr(f)
        _ = str(f)
        _ = f.dump(display=False)
        self.assertTrue(f.construct_type == 'field')


    def test_Field___getitem__(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f.copy()
        f = f.squeeze()
        
        d = f.data.array
        
        g = f[...]
        self.assertTrue((g.data.array == d).all())
        
        g = f[:, :]
        self.assertTrue((g.data.array == d).all())
        
        g = f[slice(None), :]
        self.assertTrue((g.data.array == d).all())
        
        g = f[slice(None), slice(None)]
        self.assertTrue((g.data.array == d).all())
        
        g = f[slice(None), ...]
        self.assertTrue((g.data.array == d).all())
        
        g = f[..., slice(None)]
        self.assertTrue((g.data.array == d).all())
        
        g = f[:, slice(0, f.data.shape[1], 1)]
        self.assertTrue((g.data.array == d).all())
        
        for indices, shape, multiple_list_indices in (
                [(slice(0, None, 1), slice(0, None)) , (10, 9), False],
                [(slice(3, 7)      , slice(2, 5))    , ( 4, 3), False],
                [(slice(6, 2, -1)  , slice(4, 1, -1)), ( 4, 3), False],
#               [(1                , 3)              , ( 1, 1), False],
#               [(-2               , -4)             , ( 1, 1), False],
#               [(-2               , slice(1, 5))    , ( 1, 4), False],
#               [(slice(5, 1, -2)  , 7)              , ( 2, 1), False],
                [([1, 4, 7]        , slice(1, 5))    , ( 3, 4), False],
                [([1, 4, 7]        , slice(6, 8))    , ( 3, 2), False],
                [(slice(6, 8)      , [1, 4, 7])      , ( 2, 3), False],
                [([0, 3, 8]        , [1, 7, 8])      , ( 3, 3), True],
                [([8, 3, 0]        , [8, 7, 1])      , ( 3, 3), True]
        ):
            g = f[indices]

            if not multiple_list_indices:
                e = d[indices]
            else:
                e = d.copy()
                indices = list(indices)
                for axis, i in enumerate(indices):
                    if isinstance(i, list):
                        e = numpy.take(e, indices=i, axis=axis)
                        indices[axis] = slice(None)
    
                e = e[tuple(indices)]
            #--- End: if
            
            self.assertTrue(g.data.shape == e.data.shape,
                            'Bad shape for {}: {} != {}'.format(
                                indices,
                                g.data.shape,
                                e.data.shape))
            self.assertTrue((g.data.array == e).all(),
                            'Bad values for {}: {} != {}'.format(indices,
                                                                 g.data.array,
                                                                 e))
        #--- End: for

        # Check slicing of bounds
        g = f[..., 0:4]
        c = g.construct('grid_longitude')
        b = c.bounds
        self.assertTrue(c.data.shape == (4,))
        self.assertTrue(b.data.shape == (4, 2))


#    def test_Field___setitem__(self):
#        if self.test_only and inspect.stack()[0][3] not in self.test_only:
#            return
#
#        f = self.f.squeeze()
#        
#        f[...] = 0
#        self.assertTrue((f.data.array == 0).all())
#
#        f[:, :] = 0
#        self.assertTrue((f.data.array == 0).all())
#
#
#        for indices in [
#                (slice(None)    , slice(None)),
#                (slice(3, 7)    , slice(None)),
#                (slice(None)    , slice(2, 5)),
#                (slice(3, 7)    , slice(2, 5)),
#                (slice(6, 2, -1), slice(4, 1, -1)),
#                (slice(2, 6)    , slice(4, 1, -1)),
#                ([0, 3, 8]      , [1, 7, 8]),
#                ([7, 4, 1]      , slice(6, 8)),
#        ]:
#            f[...] = 0
#            f[indices] = -1
#            array = f[indices].data.array
#            self.assertTrue((array == -1).all())
#            
#            values, counts = numpy.unique(f.data.array, return_counts=True)
#            self.assertTrue(counts[0] == array.size)


    def test_Field_PROPERTIES(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f.copy()
        for name, value in f.properties().items():
            self.assertTrue(f.has_property(name))
            _ = f.get_property(name)
            _ = f.del_property(name)
            self.assertTrue(f.del_property(name, default=None) == None)
            self.assertTrue(f.get_property(name, default=None) == None)
            self.assertFalse(f.has_property(name))
            f.set_property(name, value)

        _ = f.clear_properties()
        f.set_properties(_)
        f.set_properties(_, copy=False)


    def test_Field_DATA(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f.copy()

        self.assertTrue(f.has_data())
        data = f.get_data()
        _ = f.del_data()
        _ = f.get_data(default=None)
        _ = f.del_data(default=None)
        self.assertFalse(f.has_data())
        _ = f.set_data(data, axes=None)
        _ = f.set_data(data, axes=None, copy=False)
        self.assertTrue(f.has_data())                

        f = self.f.copy()
        _ = f.del_data_axes()
        self.assertFalse(f.has_data_axes())
        self.assertTrue(f.del_data_axes(default=None) == None)

        f = self.f.copy()
        for key in f.constructs.filter_by_data():
            self.assertTrue(f.has_data_axes(key))
            _ = f.get_data_axes(key)
            _ = f.del_data_axes(key)
            self.assertTrue(f.del_data_axes(key, default=None) == None)
            self.assertTrue(f.get_data_axes(key, default=None) == None)
            self.assertFalse(f.has_data_axes(key))


    def test_Field_CONSTRUCTS(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return
        
        f = self.f.copy()

        _ = f.construct('latitude')
        self.assertTrue(f.construct('NOT_latitude', default=None) == None)
        self.assertTrue(f.construct(re.compile('^l'), default=None) == None)

        key = f.construct_key('latitude')
        _ = f.get_construct(key)
        self.assertTrue(f.get_construct('qwerty', default=None) == None)

        constructs = self.f.auxiliary_coordinates
        n = 3
        self.assertTrue(len(constructs) == n,
                        'Got {} auxiliary coordinate constructs, expected {}'.format(len(constructs), n))               
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.AuxiliaryCoordinate)

        constructs = self.f.cell_measures
        n = 1
        self.assertTrue(len(constructs) == n,
                        'Got {} cell measure constructs, expected {}'.format(len(constructs), n))               
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.CellMeasure)

        constructs = self.f.cell_methods
        n = 2
        self.assertTrue(len(constructs) == n,
                        'Got {} cell method constructs, expected {}'.format(len(constructs), n))
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.CellMethod)
    
        ordered = self.f.cell_methods.ordered()        
        self.assertIsInstance(ordered, collections.OrderedDict)
    
        constructs = self.f.coordinate_references
        n = 2
        self.assertTrue(len(constructs) == n,
                        'Got {} ccoordinate reference onstructs, expected {}'.format(len(constructs), n))
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.CoordinateReference)
  
        constructs = self.f.dimension_coordinates
        n = 3
        self.assertTrue(len(constructs) == n,
                        'Got {} dimension coordinate constructs, expected {}'.format(len(constructs), n))
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.DimensionCoordinate)
            
        constructs = self.f.domain_ancillaries
        n = 3
        self.assertTrue(len(constructs) == n,
                        'Got {} domain ancillary constructs, expected {}'.format(len(constructs), n))
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.DomainAncillary)

        constructs = self.f.domain_axes
        n = 3
        self.assertTrue(len(constructs) == n,
                        'Got {} domain axis constructs, expected {}'.format(len(constructs), n))
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.DomainAxis)

        constructs = self.f.field_ancillaries
        n = 3
        self.assertTrue(len(constructs) == n,
                        'Got {} field ancillary constructs, expected {}'.format(len(constructs), n))
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.FieldAncillary)

        # Domain axis key
        f = self.f
        ckey = f.construct_key('grid_latitude')
        dakey = f.get_data_axes(ckey)[0]
        self.assertTrue(f.domain_axis_key('grid_latitude') == dakey)        
        self.assertTrue(f.domain_axis_key('XXXX_latitude', default=None) == None)
        self.assertTrue(f.domain_axis_key(re.compile('^grid_'), default=None) == None)


    def test_Field_data_axes(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f.copy()

        ref = f.get_data_axes()
        
        self.assertTrue(f.get_data_axes(default=None) == ref)

        self.assertTrue(f.del_data_axes() == ref)
        self.assertTrue(f.del_data_axes(default=None) == None)        
        
        self.assertTrue(f.set_data_axes(ref) == None)
        self.assertTrue(f.get_data_axes() == ref)


    def test_Field_convert(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f.copy()

        key = f.construct_key('grid_latitude')        
        c = f.convert(key)

        self.assertTrue(c.data.ndim == 1)
        self.assertTrue(c.get_property('standard_name') == 'grid_latitude')
        self.assertTrue(len(c.dimension_coordinates) == 1)
        self.assertTrue(len(c.auxiliary_coordinates) == 1)
        self.assertTrue(len(c.cell_measures) == 0)
        self.assertTrue(len(c.coordinate_references) == 1)
        self.assertTrue(len(c.domain_ancillaries) == 0)
        self.assertTrue(len(c.field_ancillaries) == 0)
        self.assertTrue(len(c.cell_methods) == 0)

        key = f.construct_key('latitude')        
        c = f.convert(key)
        
        self.assertTrue(c.data.ndim == 2)
        self.assertTrue(c.get_property('standard_name') == 'latitude')
        self.assertTrue(len(c.dimension_coordinates) == 2)
        self.assertTrue(len(c.auxiliary_coordinates) == 3)
        self.assertTrue(len(c.cell_measures) == 1)
        self.assertTrue(len(c.coordinate_references) == 1)
        self.assertTrue(len(c.domain_ancillaries) == 0)
        self.assertTrue(len(c.field_ancillaries) == 0)
        self.assertTrue(len(c.cell_methods) == 0)

        with self.assertRaises(ValueError):
            f.convert('qwerty')


    def test_Field_equals(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f.copy()
        self.assertTrue(f.equals(f, verbose=True))
 
        g = f.copy()
        self.assertTrue(f.equals(g, verbose=True))
        self.assertTrue(g.equals(f, verbose=True))

        g = f[...]
        self.assertTrue(f.equals(g, verbose=True))
        self.assertTrue(g.equals(f, verbose=True))

        g = g.squeeze()
        self.assertFalse(f.equals(g))

        h = f.copy()
        h.data[...] = h.data.array[...] + 1
        self.assertFalse(f.equals(h))


    def test_Field_del_construct(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f.copy()

        a = f.del_construct('auxiliarycoordinate1')
        self.assertTrue(a.construct_type == 'auxiliary_coordinate')

        try:
            a = f.del_construct('auxiliarycoordinate1')
        except ValueError:
            pass
            
        a = f.del_construct('auxiliarycoordinate1', default=None)
        self.assertTrue(a == None)        

    
    def test_Field_squeeze_transpose_insert_dimension(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f

        g = f.transpose()
        self.assertTrue(g.data.shape == f.data.shape[::-1])
        self.assertTrue(g.get_data_axes() == f.get_data_axes()[::-1])

        g = f.squeeze()
        self.assertTrue(g.data.shape == f.data.shape[1:])
        self.assertTrue(g.get_data_axes() == f.get_data_axes()[1:], (g.get_data_axes(), f.get_data_axes()))
        
        f = f.copy()
        g = f.copy()

        key = g.set_construct(cfdm.DomainAxis(1))
        h = g.insert_dimension(axis=key)
        self.assertTrue(h.data.ndim == f.data.ndim + 1)
        self.assertTrue(h.get_data_axes()[1:] == f.get_data_axes())

        key = g.set_construct(cfdm.DomainAxis(1))
        h = g.insert_dimension(position=g.data.ndim, axis=key)
        self.assertTrue(h.data.ndim == f.data.ndim + 1)
        self.assertTrue(h.get_data_axes()[:-1] == f.get_data_axes())


#--- End: class

if __name__ == '__main__':
    print('Run date:', datetime.datetime.now())
    print(cfdm.environment(display=False))
    print('')
    unittest.main(verbosity=2)



from builtins import (object, str)
from future.utils import with_metaclass

import abc

from copy import copy, deepcopy


class Container(with_metaclass(abc.ABCMeta, object)):
    '''Abstract base class for storing components.

.. versionadded:: 1.7.0

    '''
    def __init__(self, source=None, copy=True):
        '''**Initialisation**

    A container is initialised with no parameters. Components are set
    after initialisation with the `_set_component` method.

        '''
        self._components = {}

        if source is not None:
            try:
                custom = source._get_component('custom', {})
            except AttributeError:
                custom = {}
            else:
                custom = custom.copy()
#                if custom:                  
#                        custom = deepcopy(custom)
#                    else:
#                        custom = custom.copy()
#                else:
#                    custom = {}
        else:
            custom = {}

        self._set_component('custom', custom, copy=False)

        
    def __deepcopy__(self, memo):
        '''Called by the `copy.deepcopy` function.

    x.__deepcopy__() <==> copy.deepcopy(x)
    
    .. versionadded:: 1.7.0
    
    **Examples:**
    
    >>> import copy
    >>> y = copy.deepcopy(x)

        '''
        return self.copy()


    # ----------------------------------------------------------------
    # Private methods
    # ----------------------------------------------------------------
    def _default(self, default, message=None):
        '''TODO

    .. versionadded:: 1.7.0

    :Parameters:
    
        default: 
            TODO
            
        message: `str`, optional 
            TODO
            
    :Returns:
    
        TODO
    
    **Examples:**
    
    TODO

        '''
        if isinstance(default, Exception):
            if message is not None and not default.args:
                default = copy(default)
                default.args = (message,)

            raise default
        
        return default

    
    def _del_component(self, component, default=ValueError()):
        '''Remove a component.

    .. versionadded:: 1.7.0
    
    .. seealso:: `_get_component`, `_has_component`, `_set_component`
    
    :Parameters:
    
        component: 
            The name of the component to be removed.
    
        default: optional
            Return *default* if the component has not been set.
    
    :Returns:
    
            The removed component. If unset then *default* is
            returned, if provided.
    
    **Examples:**
    
    >>> f._set_component('foo', 'bar')
    >>> f._has_component('foo')
    True
    >>> f._get_component('foo')
    'bar'
    >>> f._del_component('foo')
    'bar'
    >>> f._has_component('foo')
    False

        '''
        try:
            return self._components.pop(component)
        except KeyError:
             return self._default(default,
                                  "{!r} has no {!r} component".format(
                                      self.__class__.__name__, component))


    @property
    def _custom(self):
        '''Storage for additional attributes.

    .. versionadded:: 1.7.4
    
    **Examples:**
    
    >>> f._custom
    {}
    >>> f._custom['feature'] = ['f']
    >>> g = f.copy()
    >>> g._custom['feature'][0] = 'g'
    >>> f._custom
    {'feature': ['f']}
    >>> g._custom
    {'feature': ['g']}
    >>> del g._custom['feature']
    >>> f._custom
    {'feature': ['f']}
    >>> g._custom
    {}
    
    '''
        return self._get_component('custom')

    
    def _get_component(self, component, default=ValueError()):
        '''Return a component

    .. versionadded:: 1.7.0
    
    .. seealso:: `_del_component`, `_has_component`, `_set_component`
    
    :Parameters:
    
        component: 
            The name of the component to be returned.
    
        default: optional
            Return *default* if the component has not been set.
    
    :Returns:
    
            The component. If unset then *default* is returned, if
            provided.
    
    **Examples:**
    
    >>> f._set_component('foo', 'bar')
    >>> f._has_component('foo')
    True
    >>> f._get_component('foo')
    'bar'
    >>> f._del_component('foo')
    'bar'
    >>> f._has_component('foo')
    False

        '''
        try:            
            return self._components[component]
        except KeyError:
            return self._default(default,
                                 "{!r} has no {!r} component".format(
                                     self.__class__.__name__, component))


    def _has_component(self, component):
        '''Whether a component has been set.

    .. versionadded:: 1.7.0
    
    .. seealso:: `_del_component`, `_get_component`, `_set_component`
    
    :Parameters:
    
        component: 
            The name of the component.
    
    :Returns:
    
        `bool`
            True if the component has been set, otherwise False.
    
    **Examples:**
    
    >>> f._set_component('foo', 'bar')
    >>> f._has_component('foo')
    True
    >>> f._get_component('foo')
    'bar'
    >>> f._del_component('foo')
    'bar'
    >>> f._has_component('foo')
    False

        '''
        return component in self._components


    def _set_component(self, component, value, copy=True):
        '''Set a component.

    .. versionadded:: 1.7.0
    
    .. seealso:: `_del_component`, `_get_component`, `_has_component`
    
    :Parameters:
    
        component: `str`
            The name of the component.
    
        value:
            The value for the component.
    
    :Returns:
    
        `None`
    
    **Examples:**
        
    >>> f._set_component('foo', 'bar')
    >>> f._has_component('foo')
    True
    >>> f._get_component('foo')
    'bar'
    >>> f._del_component('foo')
    'bar'
    >>> f._has_component('foo')
    False

        '''
        if copy:
            value = deepcopy(value)
            
        self._components[component] = value


    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def copy(self):
        '''Return a deep copy.

    ``f.copy()`` is equivalent to ``copy.deepcopy(f)``.
    
    .. versionadded:: 1.7.0
    
    :Returns:
    
            The deep copy.
    
    **Examples:**
    
    >>> g = f.copy()

        '''
        return type(self)(source=self, copy=True)

    
#--- End: class



from __future__ import print_function
from past.builtins import basestring

import os

from . import implementation

from .netcdf import NetCDFRead


_implementation = implementation()
       

def read(filename, external=None, extra=None, verbose=False,
         warnings=False, _implementation=_implementation):
    '''Read field constructs from a dataset.

    The dataset may be a netCDF file on disk or on an OPeNDAP server,
    or a CDL file on disk (see below).
    
    The returned field constructs are sorted by the netCDF variable
    names of their corresponding data variables.
    
    **CDL files**

    A file is considered to be a CDL (Common Data Language)
    representation of a netCDF dataset if it is a text file that
    starts with the seven characters "netcdf " (six letters followed
    by a space). It is converted to a temporary netCDF4 file using the
    external ``ncgen`` command, and the temporary file persists until
    the end of the Python session, at which time it is automatically
    deleted. The CDL file may omit the data array values (as would be
    the case, for example, if the file was created with the ``-h`` or
    ``-c`` option to ``ncdump``), in which case the the relevant
    constructs in memory will be created with data containing missing
    values.
   
    **NetCDF unlimited dimensions**
    
    Domain axis constructs that correspond to NetCDF unlimited
    dimensions may be accessed with the
    `~cf.DomainAxis.nc_is_unlimited` and
    `~cf.DomainAxis.nc_set_unlimited` methods of a domain axis
    construct.

    **CF-compliance**
    
    If the dataset is partially CF-compliant to the extent that it is
    not possible to unambiguously map an element of the netCDF dataset
    to an element of the CF data model, then a field construct is
    still returned, but may be incomplete. This is so that datasets
    which are partially conformant may nonetheless be modified in
    memory and written to new datasets.
    
    Such "structural" non-compliance would occur, for example, if the
    "coordinates" attribute of a CF-netCDF data variable refers to
    another variable that does not exist, or refers to a variable that
    spans a netCDF dimension that does not apply to the data
    variable. Other types of non-compliance are not checked, such
    whether or not controlled vocabularies have been adhered to. The
    structural compliance of the dataset may be checked with the
    `~cfdm.Field.dataset_compliance` method of the field construct, as
    well as optionally displayed when the dataset is read by setting
    the *warnings* parameter.
    
    **Performance**
    
    Descriptive properties are always read into memory, but lazy
    loading is employed for all data arrays, which means that no data
    is read into memory until the data is required for inspection or
    to modify the array contents. This maximises the number of field
    constructs that may be read within a session, and makes the read
    operation fast.

    .. versionadded:: 1.7.0
    
    .. seealso:: `cfdm.write`, `cfdm.Field.convert`,
                 `cfdm.Field.dataset_compliance`
    
    :Parameters:
    
        filename: `str`
            The file name or OPenDAP URL of the dataset.
    
            Relative paths are allowed, and standard tilde and shell
            parameter expansions are applied to the string.
    
            *Parameter example:*
              The file ``file.nc`` in the user's home directory could
              be described by any of the following:
              ``'$HOME/file.nc'``, ``'${HOME}/file.nc'``,
              ``'~/file.nc'``, ``'~/tmp/../file.nc'``.
        
        external: (sequence of) `str`, optional
            Read external variables (i.e. variables which are named by
            attributes, but are not present, in the parent file given
            by the *filename* parameter) from the given external
            files. Ignored if the parent file does not contain a
            global "external_variables" attribute. Multiple external
            files may be provided, which are searched in random order
            for the required external variables.
           
            If an external variable is not found in any external
            files, or is found in multiple external files, then the
            relevant metadata construct is still created, but without
            any metadata or data. In this case the construct's
            `!is_external` method will return `True`.
    
            *Parameter example:*
              ``external='cell_measure.nc'``
    
            *Parameter example:*
              ``external=['cell_measure.nc']``
    
            *Parameter example:*
              ``external=('cell_measure_A.nc', 'cell_measure_O.nc')``
    
        extra: (sequence of) `str`, optional
            Create extra, independent fields from netCDF variables
            that correspond to particular types metadata
            constructs. The *extra* parameter may be one, or a
            sequence, of:
    
              ==========================  ================================
              *extra*                     Metadata constructs
              ==========================  ================================
              ``'field_ancillary'``       Field ancillary constructs
              ``'domain_ancillary'``      Domain ancillary constructs
              ``'dimension_coordinate'``  Dimension coordinate constructs
              ``'auxiliary_coordinate'``  Auxiliary coordinate constructs
              ``'cell_measure'``          Cell measure constructs
              ==========================  ================================
    
            *Parameter example:*
              To create fields from auxiliary coordinate constructs:
              ``extra='auxiliary_coordinate'`` or
              ``extra=['auxiliary_coordinate']``.
    
            *Parameter example:*
              To create fields from domain ancillary and cell measure
              constructs: ``extra=['domain_ancillary',
              'cell_measure']``.
    
            An extra field construct created via the *extra* parameter
            will have a domain limited to that which can be inferred
            from the corresponding netCDF variable, but without the
            connections that are defined by the parent netCDF data
            variable. It is possible to create independent fields from
            metadata constructs that do incorporate as much of the
            parent field construct's domain as possible by using the
            `~cfdm.Field.convert` method of a returned field
            construct, instead of setting the *extra* parameter.
    
        verbose: `bool`, optional
            If True then print a description of how the contents of
            the netCDF file were parsed and mapped to CF data model
            constructs.
    
        warnings: `bool`, optional
            If True then print warnings when an output field construct
            is incomplete due to structural non-compliance of the
            dataset. By default such warnings are not displayed.
            
        _implementation: (subclass of) `CFDMImplementation`, optional
            Define the CF data model implementation that provides the
            returned field constructs.
    
    :Returns:
        
        `list`
            The field constructs found in the dataset. The list may be
            empty.
    
    **Examples:**
    
    >>> x = cfdm.rea

    d('file.nc')
    >>> print(type(x))
    <type 'list'>
    
    Read a file and create field constructs from CF-netCDF data
    variables as well as from the netCDF variables that correspond to
    particular types metadata constructs:
    
    >>> f = cfdm.read('file.nc', extra='domain_ancillary')
    >>> g = cfdm.read('file.nc', extra=['dimension_coordinate', 
    ...                                 'auxiliary_coordinate'])
    
    Read a file that contains external variables:
    
    >>> h = cfdm.read('parent.nc')
    >>> i = cfdm.read('parent.nc', external='external.nc')
    >>> j = cfdm.read('parent.nc', external=['external1.nc', 'external2.nc'])

    '''
    # Parse the field parameter
    if extra is None:
        extra = ()
    elif isinstance(extra, basestring):
        extra = (extra,)

    filename = os.path.expanduser(os.path.expandvars(filename))
    
    if os.path.isdir(filename):
        raise IOError("Can't read directory {}".format(filename))

    if not os.path.isfile(filename):
        raise IOError("Can't read non-existent file {}".format(filename))

    # ----------------------------------------------------------------
    # Read the fields in the file
    # ----------------------------------------------------------------

    # Initialise a netCDF read object
    netcdf = NetCDFRead(_implementation)

    # Read the file into fields.
    cdl = False
    if netcdf.is_cdl_file(filename):        
        # Create a temporary netCDF file from input CDL
        cdl = True
        cdl_filename = filename
        filename = netcdf.cdl_to_netcdf(filename)

    if netcdf.is_netcdf_file(filename):
        fields = netcdf.read(filename, external=external, extra=extra,
                             verbose=verbose, warnings=warnings,
                             extra_read_vars=None)
    elif cdl:
        raise IOError(
            "Can't determine format of file {} generated from CDL file {}".format(
                filename, cdl_filename))
    else:
        raise IOError("Can't determine format of file {}".format(filename))

    # ----------------------------------------------------------------
    # Return the fields
    # ----------------------------------------------------------------
    return fields



'''

'''

__author__       = 'David Hassell'
__date__         = '2019-11-07'
__cf_version__   = '1.7'
#__cf_version__   = '1.8'
__version__      = '1.7.9'

from distutils.version import LooseVersion
import platform

_requires = ('numpy',
             'netCDF4',
             'future')

_error0 = 'cfdm.core requires the modules {}. '.format(', '.join(_requires))

try:
    import netCDF4
except ImportError as error1:
    raise ImportError(_error0+str(error1))

try:
    import numpy
except ImportError as error1:
    raise ImportError(_error0+str(error1))

try:
    import future
except ImportError as error1:
    raise ImportError(_error0+str(error1))

# Check the version of python
_minimum_vn = '2.7.0'
if LooseVersion(platform.python_version()) < LooseVersion(_minimum_vn):
    raise ValueError(
        "Bad python version: cfdm.core requires python version {} or later. Got {}".format(
            _minimum_vn,  platform.python_version()))

# Check the version of netCDF4
minimum_vn = '1.4.0'
if LooseVersion(netCDF4.__version__) < LooseVersion(minimum_vn):
    raise ValueError(
        "Bad netCDF4 version: cfdm.core requires netCDF4 version {} or later. Got {} at {}".format(
            minimum_vn, netCDF4.__version__, netCDF4.__file__))

# Check the version of numpy
minimum_vn = '1.15'
if LooseVersion(numpy.__version__) < LooseVersion(minimum_vn):
    raise ValueError(
        "Bad numpy version: cfdm.core requires numpy version {} or later. Got {} at {}".format(
            minimum_vn, numpy.__version__, numpy.__file__))

from .constructs import Constructs

from .functions import (CF,
                        environment)

from .data import (Data,
                   Array,
                   NumpyArray)

from .bounds               import Bounds
from .coordinateconversion import CoordinateConversion
from .datum                import Datum
from .domain               import Domain
from .interiorring         import InteriorRing

from .auxiliarycoordinate import AuxiliaryCoordinate
from .cellmeasure         import CellMeasure
from .cellmethod          import CellMethod
from .coordinatereference import CoordinateReference
from .dimensioncoordinate import DimensionCoordinate
from .domainancillary     import DomainAncillary
from .domainaxis          import DomainAxis
from .field               import Field
from .fieldancillary      import FieldAncillary



from __future__ import print_function

import datetime
import os
import sys
import unittest

import numpy

import cfdm

verbose  = False
warnings = False

class create_fieldTest_2(unittest.TestCase):
    filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                            'test_file_b.nc')

    def test_create_field_2(self):

        # Dimension coordinates
        dim1 = cfdm.DimensionCoordinate(data=cfdm.Data(numpy.arange(10.)))
        dim1.set_property('standard_name', 'projection_y_coordinate')
        dim1.set_property('units', 'm')

        data = numpy.arange(9.) + 20
        data[-1] = 34
        dim0 = cfdm.DimensionCoordinate(data=cfdm.Data(data))
        dim0.set_property('standard_name', 'projection_x_coordinate')
        dim0.set_property('units', 'm')

        array = dim0.data.array

        array = numpy.array([array-0.5, array+0.5]).transpose((1,0))
        array[-2, 1] = 30
        array[-1, :] = [30, 36]
        dim0.set_bounds(cfdm.Bounds(data=cfdm.Data(array)))
        
        dim2 = cfdm.DimensionCoordinate(data=cfdm.Data([1.5]),
                                        bounds=cfdm.Bounds(data=cfdm.Data([[1, 2.]])))
        dim2.set_property('standard_name', 'atmosphere_hybrid_height_coordinate')
        
        # Auxiliary coordinates
        aux2 = cfdm.AuxiliaryCoordinate(
            data=cfdm.Data(numpy.arange(-45, 45, dtype='int32').reshape(10, 9)))
        aux2.set_property('units', 'degree_N')
        aux2.set_property('standard_name', 'latitude')
        
        aux3 = cfdm.AuxiliaryCoordinate(
            data=cfdm.Data(numpy.arange(60, 150, dtype='int32').reshape(9, 10)))
        aux3.set_property('standard_name', 'longitude')
        aux3.set_property('units', 'degreeE')

        array = numpy.ma.array(['alpha','beta','gamma','delta','epsilon',
                                'zeta','eta','theta','iota','kappa'], dtype='S')
        array[0] = numpy.ma.masked
        aux4 = cfdm.AuxiliaryCoordinate(data=cfdm.Data(array))
        aux4.set_property('standard_name', 'greek_letters')

        # Domain ancillaries
        ak = cfdm.DomainAncillary(data=cfdm.Data([10.]))
        ak.set_property('units', 'm')
        ak.set_bounds(cfdm.Bounds(data=cfdm.Data([[5, 15.]])))
        
        bk = cfdm.DomainAncillary(data=cfdm.Data([20.]))
        bk.set_bounds(cfdm.Bounds(data=cfdm.Data([[14, 26.]])))
        
        # Cell measures
        msr0 = cfdm.CellMeasure(
            data=cfdm.Data(1+numpy.arange(90.).reshape(9, 10)*1234))
        msr0.set_measure('area')
        msr0.set_property('units', 'km2')
        
        # Data          
        data = cfdm.Data(numpy.arange(90.).reshape(10, 9))

        properties = {'units': 'm s-1'}
        
        f = cfdm.Field(properties=properties)
        f.set_property('standard_name', 'eastward_wind')

        axisX = f.set_construct(cfdm.DomainAxis(9))
        axisY = f.set_construct(cfdm.DomainAxis(10))
        axisZ = f.set_construct(cfdm.DomainAxis(1))

        f.set_data(data, axes=[axisY, axisX])
        
        x = f.set_construct(dim0, axes=[axisX])
        y = f.set_construct(dim1, axes=[axisY])
        z = f.set_construct(dim2, axes=[axisZ])

        lat   = f.set_construct(aux2, axes=[axisY, axisX])
        lon   = f.set_construct(aux3, axes=[axisX, axisY])
        greek = f.set_construct(aux4, axes=[axisY])

        ak = f.set_construct(ak, axes=[axisZ])
        bk = f.set_construct(bk, axes=[axisZ])

        # Coordinate references
        coordinate_conversion = cfdm.CoordinateConversion(
            parameters={'grid_mapping_name':  "transverse_mercator",
                        'latitude_of_projection_origin': 49.0,
                        'longitude_of_central_meridian': -2.0,
                        'scale_factor_at_central_meridian':  0.9996012717,
                        'false_easting': 400000.0,
                        'false_northing': -100000.0,
                        'unit': "metre"})

        datum0 = cfdm.Datum(parameters={'inverse_flattening': 299.3249646,
                                        'longitude_of_prime_meridian':  0.0,
                                        'semi_major_axis': 6377563.396})
        
        ref0 = cfdm.CoordinateReference(coordinates=[x, y],
                                        datum=datum0,
                                        coordinate_conversion=coordinate_conversion)

        coordinate_conversion = cfdm.CoordinateConversion(
            parameters={'grid_mapping_name': "latitude_longitude"})
        
        datum2 = cfdm.Datum(parameters={'longitude_of_prime_meridian': 0.0,
                                        'semi_major_axis': 6378137.0,
                                        'inverse_flattening': 298.257223563})
        
        ref2 = cfdm.CoordinateReference(
            coordinates=[lat, lon],
            datum=datum2,
            coordinate_conversion=coordinate_conversion)

        f.set_construct(msr0, axes=[axisX, axisY])

        f.set_construct(ref0)
        f.set_construct(ref2)

        orog = cfdm.DomainAncillary(data=f.get_data())
        orog.set_property('standard_name', 'surface_altitude')
        orog.set_property('units', 'm')
        orog = f.set_construct(orog, axes=[axisY, axisX])

        coordinate_conversion = cfdm.CoordinateConversion(
            parameters={'standard_name': 'atmosphere_hybrid_height_coordinate'},
            domain_ancillaries={'orog': orog,
                                'a'   : ak,
                                'b'   : bk})
        
        ref1 = cfdm.CoordinateReference(coordinates=[z],
                                        datum=datum0,
                                        coordinate_conversion=coordinate_conversion)
        
        f.set_construct(ref1)
        
        # Field ancillary variables
        g = f.copy()
        anc = cfdm.FieldAncillary(data=g.get_data())
        anc.standard_name = 'ancillaryA'
        f.set_construct(anc, axes=[axisY, axisX])
        
        g = f[0]
        g = g.squeeze()
        anc = cfdm.FieldAncillary(data=g.get_data())
        anc.standard_name = 'ancillaryB'
        f.set_construct(anc, axes=[axisX])

        g = f[..., 0]
        g = g.squeeze()
        anc = cfdm.FieldAncillary(data=g.get_data())
        anc.standard_name = 'ancillaryC'
        f.set_construct(anc, axes=[axisY])

        
        f.set_property('flag_values', numpy.array([1, 2, 4], 'int32'))
        f.set_property('flag_meanings', 'a bb ccc')
        f.set_property('flag_masks', [2, 1, 0])

        cm0 =  cfdm.CellMethod(axes=[axisX],
                               method='mean',
                               qualifiers={'interval': [cfdm.Data(1, 'day')],
                                           'comment' : 'ok'})
    
        cm1 =  cfdm.CellMethod(axes=[axisY],
                               method='maximum',                               
                               qualifiers={'where' : 'sea'})

        f.set_construct(cm0)
        f.set_construct(cm1)

        if verbose:                
            print(repr(f))
            print(f)
            print(f.constructs)
            print(f.construct_data_axes())
        
        self.assertTrue(f.equals(f.copy(), verbose=True),
                        "Field f not equal to a copy of itself")
        
#        f.dump()
        if verbose:
            print("####################################################")
            
        for fmt in ('NETCDF3_CLASSIC',
                    'NETCDF3_64BIT',
                    'NETCDF4',
                    'NETCDF4_CLASSIC'):
            cfdm.write(f, self.filename, fmt=fmt, verbose=verbose)

            g = cfdm.read(self.filename, verbose=verbose)
            if verbose:
                for x in g:
                    x.print_read_report()
        
            self.assertTrue(len(g) == 1, '{} != 1'.format(len(g)))

            g = g[0].squeeze()

            #            g[0].dump()
            
            #        g.dump
            if verbose:
                print('f')
                print(f)
                print('g')
                print(g)
                
            self.assertTrue(sorted(f.constructs) == sorted(g.constructs),
                            '\n\nf\n{}\n\n{}\n\ng\n{}\n\n{}'.format(
                            sorted(f.constructs),
                            sorted(f.constructs.items()),
                            sorted(g.constructs),
                            sorted(g.constructs.items())))


            self.assertTrue(g.equals(g.copy(), verbose=True),
                            "Field g not equal to a copy of itself")
            if verbose:
                print('f')
                f.dump()
                print('g')
                g.dump()
            
            self.assertTrue(g.equals(f, verbose=True),                            
                            "Field not equal to itself read back in")
        #--- End: for
        
        x = g.dump(display=False)
        x = f.dump(display=False)

        g = cfdm.read(self.filename, verbose=verbose,
                      extra=['domain_ancillary'], warnings=warnings)
        if verbose:
            for x in g:
                x.print_read_report()

            print(g)
#        for x in g:
#            x.dump()
#        h = g.field('domainancillary2')
#        h.dump()
#        print h
#        
#
#        h = g.field('domainancillary1')
#        print h
#        
#        h = g.field('domainancillary0')
#        print h
#        
#        h = g.field('cellmeasure0')
#        print h

        
    #--- End: def

#--- End: class

if __name__ == "__main__":
    print('Run date:', datetime.datetime.now())
    print(cfdm.environment(display=False))
    print('')
    unittest.main(verbosity=2)



from future.utils import with_metaclass
from builtins import super

import abc

from . import Properties


class PropertiesData(with_metaclass(abc.ABCMeta, Properties)):
    '''Abstract base class for a data array with descriptive properties.

.. versionadded:: 1.7.0

    '''
    def __init__(self, properties=None, data=None, source=None,
                 copy=True, _use_data=True):
        '''**Initialization**

:Parameters:

    properties: `dict`, optional
        Set descriptive properties. The dictionary keys are property
        names, with corresponding values. Ignored if the *source*
        parameter is set.

        *Parameter example:*
          ``properties={'standard_name': 'altitude'}``
        
        Properties may also be set after initialisation with the
        `set_properties` and `set_property` methods.

    data: `Data`, optional
        Set the data. Ignored if the *source* parameter is set.
        
        The data also may be set after initialisation with the
        `set_data` method.
        
    source: optional
        Initialize the properties and data from those of *source*.

    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__(properties=properties, source=source,
                         copy=copy)

        if source is not None:
            if not _use_data:
                data = None
            else:
                try:
                    data = source.get_data(None)
                except AttributeError:
                    data = None
        #--- End: if

        if _use_data and data is not None:
            self.set_data(data, copy=copy)
    #--- End: def

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def data(self):
        '''Return the data.

``f.data`` is equivalent to ``f.get_data()`` 

Note that a `Data` instance is returned. Use its `array` attribute to
return the data as a `numpy` array.

The units, calendar and fill value properties are, if set, inserted
into the data.

.. versionadded:: 1.7.0

.. seealso:: `Data.array`, `del_data`, `get_data`, `has_data`,
             `set_data`

:Returns:

    `Data`
        The data.

**Examples:**

>>> d = cfdm.Data(range(10))
>>> f.set_data(d)
>>> f.has_data()
True
>>> d = f.data
>>> d
<Data(10): [0, ..., 9]>
>>> f.data.shape
(10,)

        '''
        return self.get_data()
    #--- End: def

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def copy(self, data=True):
        '''Return a deep copy.

``f.copy()`` is equivalent to ``copy.deepcopy(f)``.

Arrays within `Data` instances are copied with a copy-on-write
technique. This means that a copy takes up very little extra memory,
even when the original contains very large data arrays, and the copy
operation is fast.

.. versionadded:: 1.7.0

:Parameters:

    data: `bool`, optional
        If False then do not copy data. By default data are copied.

:Returns:

        The deep copy.

**Examples:**

>>> g = f.copy()
>>> g = f.copy(data=False)
>>> g.has_data()
False

        '''
        return type(self)(source=self, copy=True, _use_data=data)
    #--- End: def

    def del_data(self, default=ValueError()):
        '''Remove the data.

.. versionadded:: 1.7.0

.. seealso:: `data`, `get_data`, `has_data`, `set_data`

:Parameters:

    default: optional
        Return the value of the *default* parameter if data have not
        been set. If set to an `Exception` instance then it will be
        raised instead.

:Returns: 

        The removed data.

**Examples:**

>>> d = cfdm.Data(range(10))
>>> f.set_data(d)
>>> f.has_data()
True
>>> f.get_data()
<Data(10): [0, ..., 9]>
>>> f.del_data()
<Data(10): [0, ..., 9]>
>>> f.has_data()
False
>>> print(f.get_data(None))
None
>>> print(f.del_data(None))
None

        '''
        data = self.get_data(None)
        self._del_component('data', default=default)
        return data
    #--- End: def

    def get_data(self, default=ValueError(), _units=True,
                 _fill_value=True):
        '''Return the data.o

Note that a `Data` instance is returned. Use its `array` attribute to
return the data as an independent `numpy` array.

The units, calendar and fill value properties are, if set, inserted
into the data.

.. versionadded:: 1.7.0

.. seealso:: `Data.array`, `data`, `del_data`, `has_data`, `set_data`

:Parameters:

    default: optional
        Return the value of the *default* parameter if data have not
        been set. If set to an `Exception` instance then it will be
        raised instead.

:Returns:

        The data.

**Examples:**

>>> d = cfdm.Data(range(10))
>>> f.set_data(d)
>>> f.has_data()
True
>>> f.get_data()
<Data(10): [0, ..., 9]>
>>> f.del_data()
<Data(10): [0, ..., 9]>
>>> f.has_data()
False
>>> print(f.get_data(None))
None
>>> print(f.del_data(None))
None

        '''
        data = self._get_component('data', None)

        if data is None:
            return self._default(default,
                                 message="{!r} has no data".format(
                                     self.__class__.__name__))

        if _units:
            # Copy the parent units and calendar to the data
            units = self.get_property('units', None)
            if units is not None:
                data.set_units(units)
            else:
                data.del_units(default=None)
                
            calendar = self.get_property('calendar', None)
            if calendar is not None:                
                data.set_calendar(calendar)
            else:
                data.del_calendar(default=None)
        #--- End: if
        
        if _fill_value:
            # Copy the fill_value to the data
            fill_value = self.get_property('fill_value', None) # TODO
            if fill_value is not None:
                data.set_fill_value(fill_value)
            else:
                data.del_fill_value(default=None)
        #--- End: if
        
        return data        
    #--- End: def

    def has_bounds(self):
        '''Whether or not there are cell bounds.

This is always False.
        
.. versionadded:: 1.7.4

.. seealso:: `has_data`

:Returns:

    `bool`
        Always False. 

**Examples:**

>>> f.has_bounds()
False

        '''
        return False
    #--- End: def

    def has_data(self):
        '''Whether a data has been set.
        
.. versionadded:: 1.7.0

.. seealso:: `data`, `del_data`, `get_data`, `set_data`

:Returns:

    `bool`
        True if a data have been set, otherwise False.

**Examples:**

>>> d = cfdm.Data(range(10))
>>> f.set_data(d)
>>> f.has_data()
True
>>> f.get_data()
<Data(10): [0, ..., 9]>
>>> f.del_data()
<Data(10): [0, ..., 9]>
>>> f.has_data()
False
>>> print(f.get_data(None))
None
>>> print(f.del_data(None))
None

        '''     
        return self._has_component('data')
    #--- End: def

    def set_data(self, data, copy=True):
        '''Set the data.

The units, calendar and fill value of the incoming `Data` instance are
removed prior to insertion.

.. versionadded:: 1.7.0

.. seealso:: `data`, `del_data`, `get_data`, `has_data`

:Parameters:

    data: `Data`
        The data to be inserted.

    copy: `bool`, optional
        If False then do not copy the data prior to insertion. By
        default the data are copied.

:Returns:

    `None`

**Examples:**

>>> d = Data(range(10))
>>> f.set_data(d)
>>> f.has_data()
True
>>> f.get_data()
<Data(10): [0, ..., 9]>
>>> f.del_data()
<Data(10): [0, ..., 9]>
>>> f.has_data()
False
>>> print(f.get_data(None))
None
>>> print(f.del_data(None))
None

        '''
        if copy:
            data = data.copy()

#        data.set_units(None) # TODO
#        data.set_calendar(None) # TODO
        
        self._set_component('data', data, copy=False)
    #--- End: def

#--- End: class



from __future__ import print_function
from builtins import object

import inspect

import numpy

from ..functions import ATOL, RTOL

class Container(object):
    '''Mixin class for storing object components.

    .. versionadded:: 1.7.0

    '''
    def __repr__(self):
        '''Called by the `repr` built-in function.

    x.__repr__() <==> repr(x)

    .. versionadded:: 1.7.0

        '''
        return '<{0}: {1}>'.format(self.__class__.__name__,
                                   str(self))


    def __str__(self):
        '''Called by the `str` built-in function.

    x.__str__() <==> str(x)

    .. versionadded:: 1.7.0

        '''
        out = sorted(self._components)
        return ', '.join(out)


    @property
    def _ATOL(self):
        '''TODO
        '''
        return ATOL()

    
    @property
    def _RTOL(self):
        '''TODO
        '''
        return RTOL()

    
#    @classmethod
    def _equals(self, x, y, rtol=None, atol=None,
                ignore_data_type=False, **kwargs):
        '''Whether two objects are the same.

    Equality either uses one or other of the objects `!equals`
    methods, or casts them as numpy arrays and carried aout numericlly
    tolerant equality checks.
    
    .. versionadded:: 1.7.0
    
    :Parameters:
    
        atol: float, optional
            The tolerance on absolute differences between real
            numbers. The default value is set by the `ATOL` function.
            
        rtol: float, optional
            The tolerance on relative differences between real
            numbers. The default value is set by the `RTOL` function.

        '''
        if rtol is None:
            rtol = self._RTOL
        if atol is None:
            atol = self._ATOL

        kwargs['ignore_data_type'] = ignore_data_type
        kwargs['rtol'] = rtol
        kwargs['atol'] = atol
        
        eq = getattr(x, 'equals', None)
        if callable(eq):
            # --------------------------------------------------------
            # x has a callable "equals" method
            # --------------------------------------------------------
            # Check that the kwargs are OK
            try:
                # Python 3
                pass
#                parameters = inspect.signature(eq).bind_partial(**kwargs)
            except AttributeError:
                # Python 2
                pass
            return eq(y, **kwargs)
        
        eq = getattr(y, 'equals', None)
        if callable(eq):
            # --------------------------------------------------------
            # y has a callable "equals" method
            # --------------------------------------------------------
            # Check that the kwargs are OK
            try:
                # Python 3
                pass
#                parameters = inspect.signature(eq).bind_partial(**kwargs)
            except AttributeError:
                # Python 2
                pass
            return eq(x, **kwargs)

#        print (type(x), type(y))
#        print (repr(x), repr(y))
#        print (numpy.shape(x))
#        print (numpy.shape(y))
        if numpy.shape(x) != numpy.shape(y):
            return False

        # ------------------------------------------------------------
        # Cast x and y as numpy arrays
        # ------------------------------------------------------------
        if not isinstance(x, numpy.ndarray):
            x = numpy.asanyarray(x)
            
        if not isinstance(y, numpy.ndarray):
            y = numpy.asanyarray(y)
            
            
        # THIS IS WHERE SOME NUMPY FUTURE WARNINGS ARE COMING FROM

        if not ignore_data_type and x.dtype != y.dtype:
            if x.dtype.kind not in ('S', 'U') and y.dtype.kind not in ('S', 'U'):
                return False
        
        x_is_masked = numpy.ma.isMA(x)
        y_is_masked = numpy.ma.isMA(y)

        if not (x_is_masked or y_is_masked):
            try:            
                return bool(numpy.allclose(x, y, rtol=rtol, atol=atol))
            except (IndexError, NotImplementedError, TypeError):
                return bool(numpy.all(x == y))
        else:
            if x_is_masked and y_is_masked:
                if (x.mask != y.mask).any():
                    return False
            elif ((x_is_masked and x.mask.any()) or
                  (y_is_masked and y.mask.any())):
                return False

            try:
                return bool(numpy.ma.allclose(x, y, rtol=rtol, atol=atol))
            except (IndexError, NotImplementedError, TypeError):
                out = numpy.ma.all(x == y)
                if out is numpy.ma.masked:
                    return True
                else:
                    return bool(out)


    def _equals_preprocess(self, other, verbose=False,
                           ignore_type=False):
        '''Common preprocessing prior to testing of equality.

    * If the LHS operand is (object identity) the RHS operand then
      return True.
    
    * If ignore_type=False and the LHS operand is not of the same type, or
      a squblcass of, the RHS operand then return False
    
    * If ignore_type=True and the LHS operand is not of the same type,
      or a sublcass of, the RHS operand then instantiate a new
      instance based on the the RHS class and return it.
    
    .. versionadded:: 1.7.0

        '''
        # Check for object identity
        if self is other:
            return True

        # Check that each object is of compatible type
        if ignore_type:
            if not isinstance(other, self.__class__):
                other = type(self)(source=other, copy=False)
        elif not isinstance(other, self.__class__):
            if verbose:
                print("{0}: Incompatible type: {1}".format(
		    self.__class__.__name__, type(other)))
            return False

        return other

   
#--- End: class
